# -*- coding: utf-8 -*-

# Copyright (c) 2014, 2024, Oracle and/or its affiliates.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2.0, as
# published by the Free Software Foundation.
#
# This program is designed to work with certain software (including
# but not limited to OpenSSL) that is licensed under separate terms,
# as designated in a particular file or component or in included license
# documentation. The authors of MySQL hereby grant you an
# additional permission to link the program and your derivative works
# with the separately licensed software that they have either included with
# the program or referenced in the documentation.
#
# Without limiting anything contained in the foregoing, this file,
# which is part of MySQL Connector/Python, is also subject to the
# Universal FOSS Exception, version 1.0, a copy of which can be found at
# http://oss.oracle.com/licenses/universal-foss-exception.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License, version 2.0, for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA

"""Testing the C Extension cursors
"""

import copy
import datetime
import decimal
import logging
import sys
import unittest
import warnings

from collections import namedtuple

import tests

from mysql.connector import errorcode, errors

try:
    from _mysql_connector import MySQL, MySQLError, MySQLInterfaceError
except ImportError:
    HAVE_CMYSQL = False
else:
    HAVE_CMYSQL = True

from mysql.connector.connection_cext import CMySQLConnection
from mysql.connector.cursor_cext import (
    CMySQLCursor,
    CMySQLCursorBuffered,
    CMySQLCursorPrepared,
    CMySQLCursorPreparedDict,
    CMySQLCursorPreparedNamedTuple,
    CMySQLCursorPreparedRaw,
    CMySQLCursorRaw,
)

ARCH_64BIT = sys.maxsize > 2**32 and sys.platform != "win32"
LOGGER = logging.getLogger(tests.LOGGER_NAME)


@unittest.skipIf(HAVE_CMYSQL == False, "C Extension not available")
class CExtMySQLCursorTests(tests.CMySQLCursorTests):
    def _get_cursor(self, cnx=None):
        if not cnx:
            cnx = CMySQLConnection(**self.config)
        return CMySQLCursor(connection=cnx)

    def test___init__(self):
        self.assertRaises(errors.InterfaceError, CMySQLCursor, connection="ham")
        cur = self._get_cursor(self.cnx)
        self.assertTrue(
            hex(id(self.cnx)).upper()[2:-1] in repr(cur._connection).upper()
        )

    def test_lastrowid(self):
        cur = self._get_cursor(self.cnx)

        tbl = "test_lastrowid"
        self.setup_table(self.cnx, tbl)

        cur.execute("INSERT INTO {0} (col1) VALUES (1)".format(tbl))
        self.assertEqual(1, cur.lastrowid)

        cur.execute("INSERT INTO {0} () VALUES ()".format(tbl))
        self.assertEqual(2, cur.lastrowid)

        cur.execute("INSERT INTO {0} () VALUES (),()".format(tbl))
        self.assertEqual(3, cur.lastrowid)

        cur.execute("INSERT INTO {0} () VALUES ()".format(tbl))
        self.assertEqual(5, cur.lastrowid)

    def test__fetch_warnings(self):
        self.cnx.get_warnings = True
        cur = self._get_cursor(self.cnx)

        cur._connection = None
        self.assertRaises(errors.InterfaceError, cur._fetch_warnings)

        cur = self._get_cursor(self.cnx)

        cur.execute("SELECT 'a' + 'b'")
        cur.fetchall()
        exp = [
            ("Warning", 1292, "Truncated incorrect DOUBLE value: 'a'"),
            ("Warning", 1292, "Truncated incorrect DOUBLE value: 'b'"),
        ]
        res = cur._fetch_warnings()
        self.assertTrue(tests.cmp_result(exp, res))
        self.assertEqual(len(exp), cur._warning_count)
        self.assertEqual(len(res), cur.warning_count)

    def test_execute(self):
        self.cnx.get_warnings = True
        cur = self._get_cursor(self.cnx)

        self.assertEqual(None, cur.execute(None))

        self.assertRaises(
            errors.ProgrammingError,
            cur.execute,
            "SELECT %s,%s,%s",
            (
                "foo",
                "bar",
            ),
        )

        cur.execute("SELECT 'a' + 'b'")
        cur.fetchall()
        exp = [
            ("Warning", 1292, "Truncated incorrect DOUBLE value: 'a'"),
            ("Warning", 1292, "Truncated incorrect DOUBLE value: 'b'"),
        ]
        self.assertTrue(tests.cmp_result(exp, cur._warnings))
        self.cnx.get_warnings = False

        cur.execute("SELECT BINARY 'ham'")
        exp = [(bytearray(b"ham"),)]
        self.assertEqual(exp, cur.fetchall())
        cur.close()

        tbl = "myconnpy_cursor"
        self.setup_table(self.cnx, tbl)

        cur = self._get_cursor(self.cnx)
        stmt_insert = "INSERT INTO {0} (col1,col2) VALUES (%s,%s)".format(tbl)
        res = cur.execute(stmt_insert, (1, 100))
        self.assertEqual(None, res, "Return value of execute() is wrong.")

        stmt_select = "SELECT col1,col2 FROM {0} ORDER BY col1".format(tbl)
        cur.execute(stmt_select)
        self.assertEqual([(1, "100")], cur.fetchall(), "Insert test failed")

        data = {"id": 2}
        stmt = "SELECT col1,col2 FROM {0} WHERE col1 <= %(id)s".format(tbl)
        cur.execute(stmt, data)
        self.assertEqual([(1, "100")], cur.fetchall())

        cur.close()

    def test_executemany__errors(self):
        self.cnx.get_warnings = True
        cur = self._get_cursor(self.cnx)

        self.assertEqual(None, cur.executemany(None, []))

        cur = self._get_cursor(self.cnx)
        self.assertRaises(
            errors.ProgrammingError,
            cur.executemany,
            "programming error with string",
            "foo",
        )
        self.assertRaises(
            errors.ProgrammingError,
            cur.executemany,
            "programming error with 1 element list",
            ["foo"],
        )

        self.assertEqual(None, cur.executemany("empty params", []))
        self.assertEqual(None, cur.executemany("params is None", None))

        self.assertRaises(errors.ProgrammingError, cur.executemany, "foo", ["foo"])
        self.assertRaises(
            errors.ProgrammingError,
            cur.executemany,
            "SELECT %s",
            [("foo",), "foo"],
        )

        self.assertRaises(
            errors.ProgrammingError,
            cur.executemany,
            "INSERT INTO t1 1 %s",
            [(1,), (2,)],
        )

        cur.executemany("SELECT SHA1(%s)", [("foo",), ("bar",)])

    def test_executemany(self):
        tbl = "myconnpy_cursor"
        self.setup_table(self.cnx, tbl)

        stmt_insert = "INSERT INTO {0} (col1,col2) VALUES (%s,%s)".format(tbl)
        stmt_select = "SELECT col1,col2 FROM {0} ORDER BY col1".format(tbl)

        cur = self._get_cursor(self.cnx)

        res = cur.executemany(stmt_insert, [(1, 100), (2, 200), (3, 300)])
        self.assertEqual(3, cur.rowcount)

        res = cur.executemany("SELECT %s", [("f",), ("o",), ("o",)])
        self.assertEqual(3, cur.rowcount)

        data = [{"id": 2}, {"id": 3}]
        stmt = "SELECT * FROM {0} WHERE col1 <= %(id)s".format(tbl)
        cur.executemany(stmt, data)
        self.assertEqual(5, cur.rowcount)

        cur.execute(stmt_select)
        self.assertEqual(
            [(1, "100"), (2, "200"), (3, "300")],
            cur.fetchall(),
            "Multi insert test failed",
        )

        data = [{"id": 2}, {"id": 3}]
        stmt = "DELETE FROM {0} WHERE col1 = %(id)s".format(tbl)
        cur.executemany(stmt, data)
        self.assertEqual(2, cur.rowcount)

        stmt = "TRUNCATE TABLE {0}".format(tbl)
        cur.execute(stmt)

        stmt = (
            "/*comment*/INSERT/*comment*/INTO/*comment*/{0}(col1,col2)VALUES"
            "/*comment*/(%s,%s/*comment*/)/*comment()*/ON DUPLICATE KEY UPDATE"
            " col1 = VALUES(col1)"
        ).format(tbl)

        cur.executemany(stmt, [(4, 100), (5, 200), (6, 300)])
        self.assertEqual(3, cur.rowcount)

        cur.execute(stmt_select)
        self.assertEqual(
            [(4, "100"), (5, "200"), (6, "300")],
            cur.fetchall(),
            "Multi insert test failed",
        )

        stmt = "TRUNCATE TABLE {0}".format(tbl)
        cur.execute(stmt)

        stmt = (
            "INSERT INTO/*comment*/{0}(col1,col2)VALUES"
            "/*comment*/(%s,'/*100*/')/*comment()*/ON DUPLICATE KEY UPDATE "
            "col1 = VALUES(col1)"
        ).format(tbl)

        cur.executemany(stmt, [(4,), (5,)])
        self.assertEqual(2, cur.rowcount)

        cur.execute(stmt_select)
        self.assertEqual(
            [(4, "/*100*/"), (5, "/*100*/")],
            cur.fetchall(),
            "Multi insert test failed",
        )
        cur.close()

    def _test_callproc_setup(self, cnx):
        self._test_callproc_cleanup(cnx)
        stmt_create1 = (
            "CREATE PROCEDURE myconnpy_sp_1 "
            "(IN pFac1 INT, IN pFac2 INT, OUT pProd INT) "
            "BEGIN SET pProd := pFac1 * pFac2; END;"
        )

        stmt_create2 = (
            "CREATE PROCEDURE myconnpy_sp_2 "
            "(IN pFac1 INT, IN pFac2 INT, OUT pProd INT) "
            "BEGIN SELECT 'abc'; SELECT 'def'; SET pProd := pFac1 * pFac2; "
            "END;"
        )

        stmt_create3 = (
            "CREATE PROCEDURE myconnpy_sp_3"
            "(IN pStr1 VARCHAR(20), IN pStr2 VARCHAR(20), "
            "OUT pConCat VARCHAR(100)) "
            "BEGIN SET pConCat := CONCAT(pStr1, pStr2); END;"
        )

        stmt_create4 = (
            "CREATE PROCEDURE myconnpy_sp_4"
            "(IN pStr1 VARCHAR(20), INOUT pStr2 VARCHAR(20), "
            "OUT pConCat VARCHAR(100)) "
            "BEGIN SET pConCat := CONCAT(pStr1, pStr2); END;"
        )

        stmt_create5 = f"""
            CREATE PROCEDURE {cnx.database}.myconnpy_sp_5(IN user_value INT)
            BEGIN
                SET @user_value = user_value;
                SELECT @user_value AS 'user_value', CURRENT_TIMESTAMP as
                'timestamp';
            END
        """

        try:
            cur = cnx.cursor()
            cur.execute(stmt_create1)
            cur.execute(stmt_create2)
            cur.execute(stmt_create3)
            cur.execute(stmt_create4)
            cur.execute(stmt_create5)
        except errors.Error as err:
            self.fail("Failed setting up test stored routine; {0}".format(err))
        cur.close()

    def _test_callproc_cleanup(self, cnx):
        sp_names = (
            "myconnpy_sp_1",
            "myconnpy_sp_2",
            "myconnpy_sp_3",
            "myconnpy_sp_4",
            f"{cnx.database}.myconnpy_sp_5",
        )
        stmt_drop = "DROP PROCEDURE IF EXISTS {procname}"

        try:
            cur = cnx.cursor()
            for sp_name in sp_names:
                cur.execute(stmt_drop.format(procname=sp_name))
        except errors.Error as err:
            self.fail("Failed cleaning up test stored routine; {0}".format(err))
        cur.close()

    def test_callproc(self):
        cur = self._get_cursor(self.cnx)
        self.check_method(cur, "callproc")

        self.assertRaises(ValueError, cur.callproc, None)
        self.assertRaises(ValueError, cur.callproc, "sp1", None)

        config = tests.get_mysql_config()
        self.cnx.get_warnings = True

        self._test_callproc_setup(self.cnx)
        cur = self.cnx.cursor()

        if tests.MYSQL_VERSION < (5, 1):
            exp = ("5", "4", b"20")
        else:
            exp = (5, 4, 20)
        result = cur.callproc("myconnpy_sp_1", (exp[0], exp[1], 0))
        self.assertEqual(exp, result)

        if tests.MYSQL_VERSION < (5, 1):
            exp = ("6", "5", b"30")
        else:
            exp = (6, 5, 30)
        result = cur.callproc("myconnpy_sp_2", (exp[0], exp[1], 0))
        self.assertTrue(isinstance(cur._stored_results, list))
        self.assertEqual(exp, result)

        exp_results = [("abc",), ("def",)]
        for i, result in enumerate(cur.stored_results()):
            self.assertEqual(exp_results[i], result.fetchone())

        exp = ("ham", "spam", "hamspam")
        result = cur.callproc("myconnpy_sp_3", (exp[0], exp[1], 0))
        self.assertTrue(isinstance(cur._stored_results, list))
        self.assertEqual(exp, result)

        exp = ("ham", "spam", "hamspam")
        result = cur.callproc("myconnpy_sp_4", (exp[0], (exp[1], "CHAR"), (0, "CHAR")))
        self.assertTrue(isinstance(cur._stored_results, list))
        self.assertEqual(exp, result)

        exp = (5,)
        result = cur.callproc(f"{self.cnx.database}.myconnpy_sp_5", exp)
        self.assertTrue(isinstance(cur._stored_results, list))
        self.assertEqual(exp, result)

        cur.close()
        self._test_callproc_cleanup(self.cnx)

    def test_fetchone(self):
        cur = self._get_cursor(self.cnx)

        self.assertRaises(errors.InterfaceError, cur.fetchone)

        cur = self.cnx.cursor()
        cur.execute("SELECT BINARY 'ham'")
        exp = (bytearray(b"ham"),)
        self.assertEqual(exp, cur.fetchone())
        self.assertEqual(None, cur.fetchone())
        cur.close()

    def test_fetchmany(self):
        """MySQLCursor object fetchmany()-method"""
        cur = self._get_cursor(self.cnx)

        self.assertRaises(errors.InterfaceError, cur.fetchmany)

        tbl = "myconnpy_fetch"
        self.setup_table(self.cnx, tbl)
        stmt_insert = "INSERT INTO {table} (col1,col2) VALUES (%s,%s)".format(table=tbl)
        stmt_select = "SELECT col1,col2 FROM {table} ORDER BY col1 DESC".format(
            table=tbl
        )

        cur = self.cnx.cursor()
        nrrows = 10
        data = [(i, str(i * 100)) for i in range(1, nrrows + 1)]
        cur.executemany(stmt_insert, data)
        cur.execute(stmt_select)
        exp = [(10, "1000"), (9, "900"), (8, "800"), (7, "700")]
        rows = cur.fetchmany(4)
        self.assertTrue(
            tests.cmp_result(exp, rows), "Fetching first 4 rows test failed."
        )
        exp = [(6, "600"), (5, "500"), (4, "400")]
        rows = cur.fetchmany(3)
        self.assertTrue(
            tests.cmp_result(exp, rows), "Fetching next 3 rows test failed."
        )
        exp = [(3, "300"), (2, "200"), (1, "100")]
        rows = cur.fetchmany(3)
        self.assertTrue(
            tests.cmp_result(exp, rows), "Fetching next 3 rows test failed."
        )
        self.assertEqual([], cur.fetchmany())

        # Fetch more than we have.
        cur.execute(stmt_select)
        rows = cur.fetchmany(100)
        exp = [
            (10, "1000"),
            (9, "900"),
            (8, "800"),
            (7, "700"),
            (6, "600"),
            (5, "500"),
            (4, "400"),
            (3, "300"),
            (2, "200"),
            (1, "100"),
        ]
        self.assertTrue(
            tests.cmp_result(exp, rows), "Fetching next 3 rows test failed."
        )

        # Fetch iteratively without full batch
        cur.execute(stmt_select)
        rows = cur.fetchmany(6)
        self.assertEqual(6, len(rows))

        rows = cur.fetchmany(6)
        self.assertEqual(4, len(rows))

        rows = cur.fetchmany(6)
        self.assertEqual(0, len(rows))
        self.assertEqual([], cur.fetchmany())

        cur.close()

    def test_fetchall(self):
        cur = self._get_cursor(self.cnx)

        self.assertRaises(errors.InterfaceError, cur.fetchall)

        tbl = "myconnpy_fetch"
        self.setup_table(self.cnx, tbl)
        stmt_insert = "INSERT INTO {table} (col1,col2) VALUES (%s,%s)".format(table=tbl)
        stmt_select = "SELECT col1,col2 FROM {table} ORDER BY col1 ASC".format(
            table=tbl
        )

        cur = self.cnx.cursor()
        cur.execute("SELECT * FROM {table}".format(table=tbl))
        self.assertEqual(
            [], cur.fetchall(), "fetchall() with empty result should return []"
        )
        nrrows = 10
        data = [(i, str(i * 100)) for i in range(1, nrrows + 1)]
        cur.executemany(stmt_insert, data)
        cur.execute(stmt_select)
        self.assertTrue(
            tests.cmp_result(data, cur.fetchall()), "Fetching all rows failed."
        )
        self.assertEqual(None, cur.fetchone())
        cur.close()

    def test_raise_on_warning(self):
        self.cnx.raise_on_warnings = True
        cur = self._get_cursor(self.cnx)
        cur.execute("SELECT 'a' + 'b'")
        try:
            cur.execute("SELECT 'a' + 'b'")
            cur.fetchall()
        except errors.DatabaseError:
            pass
        else:
            self.fail("Did not get exception while raising warnings.")

    def test__str__(self):
        cur = self._get_cursor(self.cnx)
        self.assertEqual("CMySQLCursor: (Nothing executed yet)", cur.__str__())

        cur.execute("SELECT VERSION()")
        cur.fetchone()
        self.assertEqual("CMySQLCursor: SELECT VERSION()", cur.__str__())
        stmt = "SELECT VERSION(),USER(),CURRENT_TIME(),NOW(),SHA1('myconnpy')"
        cur.execute(stmt)
        cur.fetchone()
        self.assertEqual("CMySQLCursor: {0}..".format(stmt[:40]), cur.__str__())
        cur.close()

    def test_column_names(self):
        cur = self._get_cursor(self.cnx)
        stmt = "SELECT NOW() as now, 'The time' as label, 123 FROM dual"
        exp = ("now", "label", "123")
        cur.execute(stmt)
        cur.fetchone()
        self.assertEqual(exp, cur.column_names)
        cur.close()

    def test_statement(self):
        cur = CMySQLCursor(self.cnx)
        exp = "SELECT * FROM ham"
        cur._executed = exp
        self.assertEqual(exp, cur.statement)
        cur._executed = "  " + exp + "    "
        self.assertEqual(exp, cur.statement)
        cur._executed = b"SELECT * FROM ham"
        self.assertEqual(exp, cur.statement)

    def test_with_rows(self):
        cur = CMySQLCursor(self.cnx)
        self.assertFalse(cur.with_rows)
        cur._description = ("ham", "spam")
        self.assertTrue(cur.with_rows)

    def tests_nextset(self):
        cur = CMySQLCursor(self.cnx)
        stmt = "SELECT 'result', 1; SELECT 'result', 2; SELECT 'result', 3"
        cur.execute(stmt)
        self.assertEqual([("result", 1)], cur.fetchall())
        self.assertTrue(cur.nextset())
        self.assertEqual([("result", 2)], cur.fetchall())
        self.assertTrue(cur.nextset())
        self.assertEqual([("result", 3)], cur.fetchall())
        self.assertEqual(None, cur.nextset())

        tbl = "myconnpy_nextset"
        stmt = (
            "SELECT 'result', 1; INSERT INTO {0} () VALUES (); "
            "SELECT * FROM {0}".format(tbl)
        )
        self.setup_table(self.cnx, tbl)

        cur.execute(stmt)
        self.assertEqual([("result", 1)], cur.fetchall())
        try:
            cur.nextset()
        except errors.Error as exc:
            self.assertEqual(errorcode.CR_NO_RESULT_SET, exc.errno)
            self.assertEqual(1, cur._affected_rows)
        self.assertTrue(cur.nextset())
        self.assertEqual([(1, None, 0)], cur.fetchall())
        self.assertEqual(None, cur.nextset())

        cur.close()
        self.cnx.rollback()

    def test_execute_multi(self):
        tbl = "myconnpy_execute_multi"
        stmt = (
            "SELECT 'result', 1; INSERT INTO {0} () VALUES (); "
            "SELECT * FROM {0}".format(tbl)
        )
        self.setup_table(self.cnx, tbl)

        cur = CMySQLCursor(self.cnx)
        results = []
        exp = [
            ("SELECT 'result', 1", [("result", 1)]),
            ("INSERT INTO {0} () VALUES ()".format(tbl), 1, 1),
            ("SELECT * FROM {0}".format(tbl), [(1, None, 0)]),
        ]
        cur.execute(stmt, map_results=True)
        results.append(
            (cur.statement, cur.fetchall())
            if cur.with_rows
            else (cur.statement, cur._affected_rows, cur.lastrowid)
        )
        while cur.nextset():
            results.append(
                (cur.statement, cur.fetchall())
                if cur.with_rows
                else (cur.statement, cur._affected_rows, cur.lastrowid)
            )

        self.assertEqual(exp, results)

        cur.close()
        self.cnx.rollback()

        cur = self._get_cursor(self.cnx)
        cur.execute("DROP PROCEDURE IF EXISTS multi_results")
        procedure = (
            "CREATE PROCEDURE multi_results () BEGIN SELECT 1; SELECT 'ham'; END"
        )
        cur.execute(procedure)
        exp_stmt = b"CALL multi_results()"
        exp_result = [[(1,)], [("ham",)], []]
        results = []
        cur.execute(exp_stmt)
        for statement, result_set in cur.fetchsets():
            self.assertEqual(exp_stmt.decode("utf-8"), statement)
            results.append(result_set)

        self.assertEqual(exp_result, results)
        cur.execute("DROP PROCEDURE multi_results")

        cur.close()

        # Bug#34655520: Wrong MySQLCursor.statement values in the results of
        # cursor.execute(..., multi=True)
        operations = [
            'select 1; SELECT "`";',
            "SELECT '\"'; SELECT 2; select '```';",
            "select 1; select '`'; select 3; select \"'''''\";",
        ]
        control = [
            ["select 1", 'SELECT "`"'],
            ["SELECT '\"'", "SELECT 2", "select '```'"],
            ["select 1", "select '`'", "select 3", "select \"'''''\""],
        ]
        with self._get_cursor(self.cnx) as cur:
            for operation, exps in zip(operations, control):
                cur.execute(operation, map_results=True)
                for (statement, result_set), exp in zip(cur.fetchsets(), exps):
                    self.assertEqual(exp, statement)


class CExtMySQLCursorBufferedTests(tests.CMySQLCursorTests):
    def _get_cursor(self, cnx=None):
        if not cnx:
            cnx = CMySQLConnection(**self.config)
        self.cnx.buffered = True
        return CMySQLCursorBuffered(connection=cnx)

    def test___init__(self):
        self.assertRaises(errors.InterfaceError, CMySQLCursorBuffered, connection="ham")

        cur = self._get_cursor(self.cnx)
        self.assertTrue(
            hex(id(self.cnx)).upper()[2:-1] in repr(cur._connection).upper()
        )

    def test_execute(self):
        self.cnx.get_warnings = True
        cur = self._get_cursor(self.cnx)

        self.assertEqual(None, cur.execute(None, None))

        self.assertEqual(True, isinstance(cur, CMySQLCursorBuffered))

        cur.execute("SELECT 1")
        self.assertEqual((1,), cur.fetchone())

    def test_raise_on_warning(self):
        self.cnx.raise_on_warnings = True
        cur = self._get_cursor(self.cnx)
        self.assertRaises(errors.DatabaseError, cur.execute, "SELECT 'a' + 'b'")

    def test_with_rows(self):
        cur = self._get_cursor(self.cnx)
        cur.execute("SELECT 1")
        self.assertTrue(cur.with_rows)

    def test_executemany(self):
        tbl = "myconnpy_cursor"
        self.setup_table(self.cnx, tbl)

        stmt_insert = "INSERT INTO {0} (col1,col2) VALUES (%s,%s)".format(tbl)
        stmt_select = "SELECT col1,col2 FROM {0} ORDER BY col1".format(tbl)

        cur = self._get_cursor(self.cnx)

        res = cur.executemany(stmt_insert, [(1, 100), (2, 200), (3, 300)])
        self.assertEqual(3, cur.rowcount)

        res = cur.executemany("SELECT %s", [("f",), ("o",), ("o",)])
        self.assertEqual(3, cur.rowcount)

        data = [{"id": 2}, {"id": 3}]
        stmt = "SELECT * FROM {0} WHERE col1 <= %(id)s".format(tbl)
        cur.executemany(stmt, data)
        self.assertEqual(5, cur.rowcount)

        cur.execute(stmt_select)
        self.assertEqual(
            [(1, "100"), (2, "200"), (3, "300")],
            cur.fetchall(),
            "Multi insert test failed",
        )

        data = [{"id": 2}, {"id": 3}]
        stmt = "DELETE FROM {0} WHERE col1 = %(id)s".format(tbl)
        cur.executemany(stmt, data)
        self.assertEqual(2, cur.rowcount)

        stmt = "TRUNCATE TABLE {0}".format(tbl)
        cur.execute(stmt)

        stmt = (
            "/*comment*/INSERT/*comment*/INTO/*comment*/{0}(col1,col2)VALUES"
            "/*comment*/(%s,%s/*comment*/)/*comment()*/ON DUPLICATE KEY UPDATE"
            " col1 = VALUES(col1)"
        ).format(tbl)

        cur.executemany(stmt, [(4, 100), (5, 200), (6, 300)])
        self.assertEqual(3, cur.rowcount)

        cur.execute(stmt_select)
        self.assertEqual(
            [(4, "100"), (5, "200"), (6, "300")],
            cur.fetchall(),
            "Multi insert test failed",
        )

        stmt = "TRUNCATE TABLE {0}".format(tbl)
        cur.execute(stmt)

        stmt = (
            "INSERT INTO/*comment*/{0}(col1,col2)VALUES"
            "/*comment*/(%s,'/*100*/')/*comment()*/ON DUPLICATE KEY UPDATE "
            "col1 = VALUES(col1)"
        ).format(tbl)

        cur.executemany(stmt, [(4,), (5,)])
        self.assertEqual(2, cur.rowcount)

        cur.execute(stmt_select)
        self.assertEqual(
            [(4, "/*100*/"), (5, "/*100*/")],
            cur.fetchall(),
            "Multi insert test failed",
        )

        # BugOra21529893
        table_name = "BugOra21529893"
        data_list = [
            [
                ("A", "B"),
            ],
            [
                ("C", "D"),
                ("O", "P"),
            ],
        ]
        stmt = f"select %s,%s from {table_name} a,{table_name} b"
        cur.execute(f"DROP TABLE IF EXISTS {table_name}")
        cur.execute(
            f"create table {table_name}"
            "(id int auto_increment primary key, c1 char(32),c2 longblob)"
            "DEFAULT CHARSET utf8"
        )
        cur.execute(
            f"insert into {table_name} (c1, c2) values('1','1'),('2','2'),('3','3')"
        )
        for data in data_list:
            cur.executemany(stmt, data)
            self.assertEqual(9 * len(data), cur.rowcount)
            self.assertEqual([data[-1] for _ in range(9)], cur.fetchall())
            self.assertEqual(9 * len(data), cur.rowcount)
            self.assertEqual([], cur.fetchall())
        cur.execute(f"DROP TABLE IF EXISTS {table_name}")

        cur.close()


class CMySQLCursorRawTests(tests.CMySQLCursorTests):
    def _get_cursor(self, cnx=None):
        if not cnx:
            cnx = CMySQLConnection(**self.config)
        return CMySQLCursorRaw(connection=cnx)

    def test_fetchone(self):
        cur = self._get_cursor(self.cnx)
        self.assertRaises(errors.InterfaceError, cur.fetchone)

        cur.execute("SELECT 1, 'string', MAKEDATE(2010,365), 2.5")
        exp = (b"1", b"string", b"2010-12-31", b"2.5")
        self.assertEqual(exp, cur.fetchone())


class CMySQLCursorPreparedTests(tests.CMySQLCursorTests):
    tbl = "prep_stmt"

    create_table_stmt = (
        "CREATE TABLE {0} ("
        "id INT AUTO_INCREMENT PRIMARY KEY, "
        "my_null INT, "
        "my_bit BIT(7), "
        "my_tinyint TINYINT, "
        "my_smallint SMALLINT, "
        "my_mediumint MEDIUMINT, "
        "my_int INT, "
        "my_bigint BIGINT, "
        "my_decimal DECIMAL(20,10), "
        "my_float FLOAT, "
        "my_double DOUBLE, "
        "my_date DATE, "
        "my_time TIME, "
        "my_datetime DATETIME, "
        "my_year YEAR, "
        "my_char CHAR(100), "
        "my_varchar VARCHAR(100), "
        "my_enum ENUM('x-small', 'small', 'medium', 'large', 'x-large'), "
        "my_geometry POINT, "
        "my_blob BLOB)"
    )

    insert_stmt = (
        "INSERT INTO {0} ("
        "my_null, "
        "my_bit, "
        "my_tinyint, "
        "my_smallint, "
        "my_mediumint, "
        "my_int, "
        "my_bigint, "
        "my_decimal, "
        "my_float, "
        "my_double, "
        "my_date, "
        "my_time, "
        "my_datetime, "
        "my_year, "
        "my_char, "
        "my_varchar, "
        "my_enum, "
        "my_geometry, "
        "my_blob) "
        "VALUES (?, B'1111100', ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, "
        "POINT(21.2, 34.2), ?)"
    )

    insert_dict_stmt = (
        "INSERT INTO {0} ("
        "my_null, "
        "my_bit, "
        "my_tinyint, "
        "my_smallint, "
        "my_mediumint, "
        "my_int, "
        "my_bigint, "
        "my_decimal, "
        "my_float, "
        "my_double, "
        "my_date, "
        "my_time, "
        "my_datetime, "
        "my_year, "
        "my_char, "
        "my_varchar, "
        "my_enum, "
        "my_geometry, "
        "my_blob) "
        "VALUES ("
        "%(my_null)s, "
        "B'1111100', "
        "%(my_tinyint)s, "
        "%(my_smallint)s, "
        "%(my_mediumint)s, "
        "%(my_int)s, "
        "%(my_bigint)s, "
        "%(my_decimal)s, "
        "%(my_float)s, "
        "%(my_double)s, "
        "%(my_date)s, "
        "%(my_time)s, "
        "%(my_datetime)s, "
        "%(my_year)s, "
        "%(my_char)s, "
        "%(my_varchar)s, "
        "%(my_enum)s, "
        "POINT(21.2, 34.2), "
        "%(my_blob)s)"
    )

    insert_columns = (
        "my_null",
        "my_tinyint",
        "my_smallint",
        "my_mediumint",
        "my_int",
        "my_bigint",
        "my_decimal",
        "my_float",
        "my_double",
        "my_date",
        "my_time",
        "my_datetime",
        "my_year",
        "my_char",
        "my_varchar",
        "my_enum",
        "my_blob",
    )

    data = (
        None,
        127,
        32767,
        8388607,
        2147483647,
        4294967295 if ARCH_64BIT else 2147483647,
        decimal.Decimal("1.2"),
        3.14,
        4.28,
        datetime.date(2018, 12, 31),
        datetime.time(12, 13, 14),
        datetime.datetime(2019, 2, 4, 10, 36, 00),
        2019,
        "abc",
        "MySQL 🐬",
        "x-large",
        bytearray(b"random blob data"),
    )

    exp = (
        None,
        124,
        127,
        32767,
        8388607,
        2147483647,
        4294967295 if ARCH_64BIT else 2147483647,
        decimal.Decimal("1.2000000000"),
        3.140000104904175,
        4.28000020980835,
        datetime.date(2018, 12, 31),
        datetime.timedelta(0, 43994),
        datetime.datetime(2019, 2, 4, 10, 36),
        2019,
        "abc",
        "MySQL \U0001f42c",
        "x-large",
        bytearray(
            b"\x00\x00\x00\x00\x01\x01\x00\x00\x003333335"
            b"@\x9a\x99\x99\x99\x99\x19A@"
        ),
        bytearray(b"random blob data"),
    )

    def setUp(self):
        config = tests.get_mysql_config()
        self.cnx = CMySQLConnection(**config)
        self.cur = self.cnx.cursor(prepared=True)
        self.cur.execute(self.create_table_stmt.format(self.tbl))

    def tearDown(self):
        self.cur.execute("DROP TABLE IF EXISTS {0}".format(self.tbl))
        self.cur.close()
        self.cnx.close()

    def test___init__(self):
        self.assertIsInstance(self.cur, CMySQLCursorPrepared)

    def test_callproc(self):
        self.assertRaises(errors.NotSupportedError, self.cur.callproc, None)

    def test_close(self):
        cur = self.cnx.cursor(prepared=True)
        self.assertEqual(None, cur._stmt)
        cur.close()

    def test_fetchone(self):
        self.cur.execute(self.insert_stmt.format(self.tbl), self.data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        row = self.cur.fetchone()
        self.assertEqual(row[1:], self.exp)
        row = self.cur.fetchone()
        self.assertIsNone(row)

    def test_fetchall(self):
        self.cur.execute(self.insert_stmt.format(self.tbl), self.data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0][1:], self.exp)

    def test_fetchmany(self):
        data = [self.data[:], self.data[:], self.data[:]]
        self.cur.executemany(self.insert_stmt.format(self.tbl), data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        rows = self.cur.fetchmany(size=2)
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0][1:], self.exp)
        self.assertEqual(rows[1][1:], self.exp)
        rows = self.cur.fetchmany(1)
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0][1:], self.exp)

    def test_execute(self):
        # Use dict as placeholders
        data_dict = dict(zip(self.insert_columns, self.data))

        self.cur.execute(self.insert_dict_stmt.format(self.tbl), data_dict)
        self.cur.execute(f"SELECT * FROM {self.tbl}")
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0][1:], self.exp)

        # Test binary string
        self.cur.execute("SELECT BINARY 'ham'")
        exp = [(bytearray(b"ham"),)]
        self.assertEqual(exp, self.cur.fetchall())

    def test_executemany(self):
        data = [self.data[:], self.data[:]]
        self.cur.executemany(self.insert_stmt.format(self.tbl), data)
        self.cur.execute(f"SELECT * FROM {self.tbl}")
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0][1:], self.exp)
        self.assertEqual(rows[1][1:], self.exp)

        # Use dict as placeholders
        data_dict = dict(zip(self.insert_columns, self.data))
        data = [data_dict, copy.deepcopy(data_dict)]

        self.cur.executemany(self.insert_dict_stmt.format(self.tbl), data)
        self.cur.execute(f"SELECT * FROM {self.tbl}")
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 4)
        self.assertEqual(rows[0][1:], self.exp)
        self.assertEqual(rows[1][1:], self.exp)


class CMySQLCursorPreparedDictTests(CMySQLCursorPreparedTests):
    tbl = "prep_dict_stmt"

    column_names = (
        "my_null",
        "my_bit",
        "my_tinyint",
        "my_smallint",
        "my_mediumint",
        "my_int",
        "my_bigint",
        "my_decimal",
        "my_float",
        "my_double",
        "my_date",
        "my_time",
        "my_datetime",
        "my_year",
        "my_char",
        "my_varchar",
        "my_enum",
        "my_geometry",
        "my_blob",
    )

    def setUp(self):
        config = tests.get_mysql_config()
        self.cnx = CMySQLConnection(**config)
        self.cur = self.cnx.cursor(prepared=True, dictionary=True)
        self.cur.execute(self.create_table_stmt.format(self.tbl))

    def tearDown(self):
        self.cur.execute("DROP TABLE IF EXISTS {0}".format(self.tbl))
        self.cur.close()
        self.cnx.close()

    def test___init__(self):
        self.assertIsInstance(self.cur, CMySQLCursorPreparedDict)

    def test_close(self):
        cur = self.cnx.cursor(prepared=True, dictionary=True)
        self.assertEqual(None, cur._stmt)
        cur.close()

    def test_fetchone(self):
        self.cur.execute(self.insert_stmt.format(self.tbl), self.data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        row = self.cur.fetchone()
        del row["id"]
        self.assertEqual(row, dict(zip(self.column_names, self.exp)))
        row = self.cur.fetchone()
        self.assertIsNone(row)

    def test_fetchall(self):
        self.cur.execute(self.insert_stmt.format(self.tbl), self.data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 1)
        del rows[0]["id"]
        self.assertEqual(rows[0], dict(zip(self.column_names, self.exp)))

    def test_fetchmany(self):
        data = [self.data[:], self.data[:], self.data[:]]
        self.cur.executemany(self.insert_stmt.format(self.tbl), data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        rows = self.cur.fetchmany(size=2)
        del rows[0]["id"]
        del rows[1]["id"]
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0], dict(zip(self.column_names, self.exp)))
        self.assertEqual(rows[1], dict(zip(self.column_names, self.exp)))

        rows = self.cur.fetchmany(1)
        del rows[0]["id"]
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0], dict(zip(self.column_names, self.exp)))

    def test_execute(self):
        # Use dict as placeholders
        data_dict = dict(zip(self.insert_columns, self.data))

        self.cur.execute(self.insert_dict_stmt.format(self.tbl), data_dict)
        self.cur.execute(f"SELECT * FROM {self.tbl}")
        rows = self.cur.fetchall()
        del rows[0]["id"]
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0], dict(zip(self.column_names, self.exp)))

    def test_executemany(self):
        data = [self.data[:], self.data[:]]
        self.cur.executemany(self.insert_stmt.format(self.tbl), data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        rows = self.cur.fetchall()
        del rows[0]["id"]
        del rows[1]["id"]
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0], dict(zip(self.column_names, self.exp)))
        self.assertEqual(rows[1], dict(zip(self.column_names, self.exp)))

        # Use dict as placeholders
        data_dict = dict(zip(self.insert_columns, self.data))
        data = [data_dict, copy.deepcopy(data_dict)]

        self.cur.executemany(self.insert_dict_stmt.format(self.tbl), data)
        self.cur.execute(f"SELECT * FROM {self.tbl}")
        rows = self.cur.fetchall()
        del rows[0]["id"]
        del rows[1]["id"]
        self.assertEqual(len(rows), 4)
        self.assertEqual(rows[0], dict(zip(self.column_names, self.exp)))
        self.assertEqual(rows[1], dict(zip(self.column_names, self.exp)))


class CMySQLCursorPreparedNamedTupleTests(CMySQLCursorPreparedTests):
    tbl = "prep_named_tuple_stmt"

    column_names = (
        "my_null",
        "my_bit",
        "my_tinyint",
        "my_smallint",
        "my_mediumint",
        "my_int",
        "my_bigint",
        "my_decimal",
        "my_float",
        "my_double",
        "my_date",
        "my_time",
        "my_datetime",
        "my_year",
        "my_char",
        "my_varchar",
        "my_enum",
        "my_geometry",
        "my_blob",
    )
    named_tuple = namedtuple("Row", column_names)

    def setUp(self):
        config = tests.get_mysql_config()
        self.cnx = CMySQLConnection(**config)
        self.cur = self.cnx.cursor(prepared=True, named_tuple=True)
        self.cur.execute(self.create_table_stmt.format(self.tbl))

    def tearDown(self):
        self.cur.execute("DROP TABLE IF EXISTS {0}".format(self.tbl))
        self.cur.close()
        self.cnx.close()

    def test___init__(self):
        self.assertIsInstance(self.cur, CMySQLCursorPreparedNamedTuple)

    def test_close(self):
        cur = self.cnx.cursor(prepared=True, named_tuple=True)
        self.assertEqual(None, cur._stmt)
        cur.close()

    def test_fetchone(self):
        self.cur.execute(self.insert_stmt.format(self.tbl), self.data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        row = self.cur.fetchone()
        self.assertEqual(row[1:], self.named_tuple(*self.exp))
        row = self.cur.fetchone()
        self.assertIsNone(row)

    def test_fetchall(self):
        data = [self.data[:], self.data[:], self.data[:]]
        self.cur.executemany(self.insert_stmt.format(self.tbl), data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 3)
        for row in rows:
            self.assertEqual(row[1:], self.named_tuple(*self.exp))

    def test_fetchmany(self):
        data = [self.data[:], self.data[:], self.data[:]]
        self.cur.executemany(self.insert_stmt.format(self.tbl), data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        rows = self.cur.fetchmany(size=2)
        self.assertEqual(len(rows), 2)
        for row in rows:
            self.assertEqual(row[1:], self.named_tuple(*self.exp))

        rows = self.cur.fetchmany(1)
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0][1:], self.named_tuple(*self.exp))

    def test_execute(self):
        # Use dict as placeholders
        data_dict = dict(zip(self.insert_columns, self.data))

        self.cur.execute(self.insert_dict_stmt.format(self.tbl), data_dict)
        self.cur.execute(f"SELECT * FROM {self.tbl}")
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0][1:], self.named_tuple(*self.exp))

    def test_executemany(self):
        data = [self.data[:], self.data[:]]
        self.cur.executemany(self.insert_stmt.format(self.tbl), data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 2)
        for row in rows:
            self.assertEqual(row[1:], self.named_tuple(*self.exp))

        # Use dict as placeholders
        data_dict = dict(zip(self.insert_columns, self.data))
        data = [data_dict, copy.deepcopy(data_dict)]

        self.cur.executemany(self.insert_dict_stmt.format(self.tbl), data)
        self.cur.execute(f"SELECT * FROM {self.tbl}")
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 4)
        for row in rows:
            self.assertEqual(row[1:], self.named_tuple(*self.exp))


class CMySQLCursorPreparedRawTests(CMySQLCursorPreparedTests):
    tbl = "prep_raw_stmt"

    column_names = (
        "my_null",
        "my_bit",
        "my_tinyint",
        "my_smallint",
        "my_mediumint",
        "my_int",
        "my_bigint",
        "my_decimal",
        "my_float",
        "my_double",
        "my_date",
        "my_time",
        "my_datetime",
        "my_year",
        "my_char",
        "my_varchar",
        "my_enum",
        "my_geometry",
        "my_blob",
    )

    def setUp(self):
        config = tests.get_mysql_config()
        self.cnx = CMySQLConnection(**config)
        self.cur = self.cnx.cursor(prepared=True, raw=True)
        self.cur.execute(self.create_table_stmt.format(self.tbl))

    def tearDown(self):
        self.cur.execute("DROP TABLE IF EXISTS {0}".format(self.tbl))
        self.cur.close()
        self.cnx.close()

    def test___init__(self):
        self.assertIsInstance(self.cur, CMySQLCursorPreparedRaw)

    def test_close(self):
        cur = self.cnx.cursor(prepared=True, raw=True)
        self.assertEqual(None, cur._stmt)
        cur.close()

    def test_fetchone(self):
        self.cur.execute(self.insert_stmt.format(self.tbl), self.data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        row = self.cur.fetchone()
        self.assertEqual(row[1:], self.exp)
        row = self.cur.fetchone()
        self.assertIsNone(row)

    def test_fetchall(self):
        data = [self.data[:], self.data[:], self.data[:]]
        self.cur.executemany(self.insert_stmt.format(self.tbl), data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 3)
        for row in rows:
            self.assertEqual(row[1:], self.exp)

    def test_fetchmany(self):
        data = [self.data[:], self.data[:], self.data[:]]
        self.cur.executemany(self.insert_stmt.format(self.tbl), data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        rows = self.cur.fetchmany(size=2)
        self.assertEqual(len(rows), 2)
        for row in rows:
            self.assertEqual(row[1:], self.exp)

        rows = self.cur.fetchmany(1)
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0][1:], self.exp)

    def test_execute(self):
        # Use dict as placeholders
        data_dict = dict(zip(self.insert_columns, self.data))

        self.cur.execute(self.insert_dict_stmt.format(self.tbl), data_dict)
        self.cur.execute(f"SELECT * FROM {self.tbl}")
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0][1:], self.exp)

    def test_executemany(self):
        data = [self.data[:], self.data[:]]
        self.cur.executemany(self.insert_stmt.format(self.tbl), data)
        self.cur.execute("SELECT * FROM {0}".format(self.tbl))
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 2)
        for row in rows:
            self.assertEqual(row[1:], self.exp)

        # Use dict as placeholders
        data_dict = dict(zip(self.insert_columns, self.data))
        data = [data_dict, copy.deepcopy(data_dict)]

        self.cur.executemany(self.insert_dict_stmt.format(self.tbl), data)
        self.cur.execute(f"SELECT * FROM {self.tbl}")
        rows = self.cur.fetchall()
        self.assertEqual(len(rows), 4)
        for row in rows:
            self.assertEqual(row[1:], self.exp)


class CMySQLCursorDeprecatedTests(tests.CMySQLCursorTests):

    def test_deprecation_cursor_prepared_raw(self):
        with warnings.catch_warnings(record=True) as warnings_stack:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always", DeprecationWarning)
            # The warning should be raised as the cursor is instantiated
            with self.cnx.cursor(prepared=True, raw=True) as _:
                pass
            self.assertTrue(
                len(warnings_stack) != 0,
                msg="No warnings were caught as warnings_stack was empty",
            )
            self.assertTrue(issubclass(warnings_stack[-1].category, DeprecationWarning))
            self.assertTrue("deprecated" in str(warnings_stack[-1].message))
            warnings.resetwarnings()

    def test_deprecation_cursor_named_tuple(self):
        with warnings.catch_warnings(record=True) as warnings_stack:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always", DeprecationWarning)
            # The warning should be raised as the cursor is instantiated
            with self.cnx.cursor(named_tuple=True) as _:
                pass
            self.assertTrue(
                len(warnings_stack) != 0,
                msg="No warnings were caught as warnings_stack was empty",
            )
            self.assertTrue(issubclass(warnings_stack[-1].category, DeprecationWarning))
            self.assertTrue("deprecated" in str(warnings_stack[-1].message))
            warnings.resetwarnings()

    def test_deprecation_cursor_buffered_named_tuple(self):
        with warnings.catch_warnings(record=True) as warnings_stack:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always", DeprecationWarning)
            # The warning should be raised as the cursor is instantiated
            with self.cnx.cursor(buffered=True, named_tuple=True) as _:
                pass
            self.assertTrue(
                len(warnings_stack) != 0,
                msg="No warnings were caught as warnings_stack was empty",
            )
            self.assertTrue(issubclass(warnings_stack[-1].category, DeprecationWarning))
            self.assertTrue("deprecated" in str(warnings_stack[-1].message))
            warnings.resetwarnings()

    def test_deprecation_cursor_prepared_named_tuple(self):
        with warnings.catch_warnings(record=True) as warnings_stack:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always", DeprecationWarning)
            # The warning should be raised as the cursor is instantiated
            with self.cnx.cursor(prepared=True, named_tuple=True) as _:
                pass
            self.assertTrue(
                len(warnings_stack) != 0,
                msg="No warnings were caught as warnings_stack was empty",
            )
            self.assertTrue(issubclass(warnings_stack[-1].category, DeprecationWarning))
            self.assertTrue("deprecated" in str(warnings_stack[-1].message))
            warnings.resetwarnings()



# Copyright (c) 2014, 2024, Oracle and/or its affiliates.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2.0, as
# published by the Free Software Foundation.
#
# This program is designed to work with certain software (including
# but not limited to OpenSSL) that is licensed under separate terms,
# as designated in a particular file or component or in included license
# documentation. The authors of MySQL hereby grant you an
# additional permission to link the program and your derivative works
# with the separately licensed software that they have either included with
# the program or referenced in the documentation.
#
# Without limiting anything contained in the foregoing, this file,
# which is part of MySQL Connector/Python, is also subject to the
# Universal FOSS Exception, version 1.0, a copy of which can be found at
# http://oss.oracle.com/licenses/universal-foss-exception.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License, version 2.0, for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA

"""Unittests for mysql.connector.django
"""

import datetime
import unittest

import tests

# Load 3rd party _after_ loading tests
try:
    from django import VERSION as DJANGO_VERSION
    from django.conf import settings
except ImportError:
    DJANGO_AVAILABLE = False
    DJANGO_VERSION = (0, 0, 0)
else:
    DJANGO_AVAILABLE = True

# Have to setup Django before loading anything else
if DJANGO_AVAILABLE:
    try:
        settings.configure()
    except RuntimeError as exc:
        if "already configured" not in str(exc):
            raise
    DBCONFIG = tests.get_mysql_config()

    settings.DATABASES = {
        "default": {
            "ENGINE": "mysql.connector.django",
            "NAME": DBCONFIG["database"],
            "USER": "root",
            "PASSWORD": DBCONFIG.get("password", ""),
            "HOST": DBCONFIG["host"],
            "PORT": DBCONFIG["port"],
            "TEST_CHARSET": "utf8",
            "TEST_COLLATION": "utf8_general_ci",
            "CONN_MAX_AGE": 0,
            "AUTOCOMMIT": True,
            "TIME_ZONE": None,
            "CONN_HEALTH_CHECKS": False,
        },
    }
    settings.SECRET_KEY = "django_tests_secret_key"
    settings.TIME_ZONE = "UTC"
    settings.USE_TZ = False
    settings.SOUTH_TESTS_MIGRATE = False
    settings.DEBUG = False

TABLES = {}
TABLES[
    "django_t1"
] = """
CREATE TABLE {table_name} (
id INT NOT NULL AUTO_INCREMENT,
c1 INT,
c2 VARCHAR(20),
INDEX (c1),
UNIQUE INDEX (c2),
PRIMARY KEY (id)
) ENGINE=InnoDB
"""

TABLES[
    "django_t2"
] = """
CREATE TABLE {table_name} (
id INT NOT NULL AUTO_INCREMENT,
id_t1 INT NOT NULL,
INDEX (id_t1),
PRIMARY KEY (id),
FOREIGN KEY (id_t1) REFERENCES django_t1(id) ON DELETE CASCADE
) ENGINE=InnoDB
"""

# Have to load django.db to make importing db backend work for Django < 1.6
import django.db  # pylint: disable=W0611

from django.core.exceptions import ImproperlyConfigured
from django.db import connection
from django.db.backends.signals import connection_created
from django.db.utils import DEFAULT_DB_ALIAS, load_backend
from django.utils.safestring import SafeText

import mysql.connector

from mysql.connector.conversion import MySQLConverter
from mysql.connector.django.introspection import FieldInfo
from mysql.connector.errors import ProgrammingError

if DJANGO_AVAILABLE:
    from mysql.connector.django.base import (
        DatabaseOperations,
        DatabaseWrapper,
        DjangoMySQLConverter,
    )
    from mysql.connector.django.introspection import DatabaseIntrospection


@unittest.skipIf(not DJANGO_AVAILABLE, "Django not available")
class DjangoSettings(tests.MySQLConnectorTests):
    """Test the Django settings."""

    def test_get_connection_params(self):
        config = tests.get_mysql_config()
        settings_dict = connection.settings_dict.copy()
        settings_dict["OPTIONS"] = {}

        # The default isolation_level should be None
        database_wrapper = DatabaseWrapper(settings_dict)
        self.assertIsNone(database_wrapper.isolation_level)

        # An invalid isolation_level should raise ImproperlyConfigured
        settings_dict["OPTIONS"]["isolation_level"] = "invalid_level"
        with self.assertRaises(ImproperlyConfigured):
            _ = DatabaseWrapper(settings_dict).get_connection_params()

        # Test a valid isolation_level
        settings_dict["OPTIONS"]["isolation_level"] = "read committed"
        database_wrapper = DatabaseWrapper(settings_dict)
        connection_params = database_wrapper.get_connection_params()
        self.assertEqual(database_wrapper.isolation_level, "read committed")
        self.assertEqual(connection_params["database"], config["database"])

        # Test session isolation level integration
        with database_wrapper.cursor() as cur:
            cur.execute("SELECT @@transaction_isolation")
            res = cur.fetchall()
            self.assertEqual(res[0][0], "READ-COMMITTED")


@unittest.skipIf(not DJANGO_AVAILABLE, "Django not available")
class DjangoIntrospection(tests.MySQLConnectorTests):
    """Test the Django introspection module"""

    cnx = None
    introspect = None

    @classmethod
    def setUpClass(cls):
        cls.cnx = DatabaseWrapper(settings.DATABASES["default"])
        cls.introspect = DatabaseIntrospection(cls.cnx)

        cur = cls.cnx.cursor()

        for table_name, sql in TABLES.items():
            cur.execute("SET foreign_key_checks = 0")
            cur.execute(
                "DROP TABLE IF EXISTS {table_name}".format(table_name=table_name)
            )
            cur.execute(sql.format(table_name=table_name))
        cur.execute("SET foreign_key_checks = 1")

    @classmethod
    def tearDownClass(cls):
        cur = cls.cnx.cursor()
        cur.execute("SET foreign_key_checks = 0")
        for table_name, sql in TABLES.items():
            cur.execute(
                "DROP TABLE IF EXISTS {table_name}".format(table_name=table_name)
            )
        cur.execute("SET foreign_key_checks = 1")

    def test_get_table_list(self):
        cur = self.cnx.cursor()
        for exp in TABLES.keys():
            res = any(
                table.name == exp for table in self.introspect.get_table_list(cur)
            )
            self.assertTrue(
                res,
                "Table {table_name} not in table list".format(table_name=exp),
            )

    def test_get_table_description(self):
        cur = self.cnx.cursor()
        if DJANGO_VERSION < (3, 2, 0):
            exp = [
                FieldInfo(
                    name="id",
                    type_code=3,
                    display_size=None,
                    internal_size=None,
                    precision=10,
                    scale=None,
                    null_ok=0,
                    default=None,
                    extra="auto_increment",
                    is_unsigned=0,
                    has_json_constraint=False,
                ),
                FieldInfo(
                    name="c1",
                    type_code=3,
                    display_size=None,
                    internal_size=None,
                    precision=10,
                    scale=None,
                    null_ok=1,
                    default=None,
                    extra="",
                    is_unsigned=0,
                    has_json_constraint=False,
                ),
                FieldInfo(
                    name="c2",
                    type_code=253,
                    display_size=None,
                    internal_size=20,
                    precision=None,
                    scale=None,
                    null_ok=1,
                    default=None,
                    extra="",
                    is_unsigned=0,
                    has_json_constraint=False,
                ),
            ]
        else:
            exp = [
                FieldInfo(
                    name="id",
                    type_code=3,
                    display_size=None,
                    internal_size=None,
                    precision=10,
                    scale=None,
                    null_ok=0,
                    default=None,
                    collation=None,
                    extra="auto_increment",
                    is_unsigned=0,
                    has_json_constraint=False,
                ),
                FieldInfo(
                    name="c1",
                    type_code=3,
                    display_size=None,
                    internal_size=None,
                    precision=10,
                    scale=None,
                    null_ok=1,
                    default=None,
                    collation=None,
                    extra="",
                    is_unsigned=0,
                    has_json_constraint=False,
                ),
                FieldInfo(
                    name="c2",
                    type_code=253,
                    display_size=None,
                    internal_size=20,
                    precision=None,
                    scale=None,
                    null_ok=1,
                    default=None,
                    collation=None,
                    extra="",
                    is_unsigned=0,
                    has_json_constraint=False,
                ),
            ]
        res = self.introspect.get_table_description(cur, "django_t1")
        self.assertEqual(exp, res)

    def test_get_relations(self):
        cur = self.cnx.cursor()
        exp = {"id_t1": ("id", "django_t1")}
        self.assertEqual(exp, self.introspect.get_relations(cur, "django_t2"))

    def test_get_key_columns(self):
        cur = self.cnx.cursor()
        exp = [("id_t1", "django_t1", "id")]
        self.assertEqual(exp, self.introspect.get_key_columns(cur, "django_t2"))

    def test_get_indexes(self):
        cur = self.cnx.cursor()
        exp = {
            "c1": {"primary_key": False, "unique": False},
            "id": {"primary_key": True, "unique": True},
            "c2": {"primary_key": False, "unique": True},
        }
        self.assertEqual(exp, self.introspect.get_indexes(cur, "django_t1"))

    def test_get_primary_key_column(self):
        cur = self.cnx.cursor()
        res = self.introspect.get_primary_key_column(cur, "django_t1")
        self.assertEqual("id", res)

    def test_get_constraints(self):
        cur = self.cnx.cursor()
        exp = {
            "PRIMARY": {
                "check": False,
                "columns": ["id"],
                "foreign_key": None,
                "index": True,
                "primary_key": True,
                "type": "idx",
                "unique": True,
            },
            "django_t2_ibfk_1": {
                "check": False,
                "columns": ["id_t1"],
                "foreign_key": ("django_t1", "id"),
                "index": False,
                "primary_key": False,
                "unique": False,
            },
            "id_t1": {
                "check": False,
                "columns": ["id_t1"],
                "foreign_key": None,
                "index": True,
                "primary_key": False,
                "type": "idx",
                "unique": False,
            },
        }
        if DJANGO_VERSION >= (3, 2, 0):
            exp["PRIMARY"]["orders"] = ["ASC"]
            exp["django_t2_ibfk_1"]["orders"] = []
            exp["id_t1"]["orders"] = ["ASC"]
        self.assertEqual(exp, self.introspect.get_constraints(cur, "django_t2"))


@unittest.skipIf(not DJANGO_AVAILABLE, "Django not available")
class DjangoDatabaseWrapper(tests.MySQLConnectorTests):
    """Test the Django base.DatabaseWrapper class"""

    def setUp(self):
        dbconfig = tests.get_mysql_config()
        self.conn = mysql.connector.connect(**dbconfig)
        self.cnx = DatabaseWrapper(settings.DATABASES["default"])

    def test__init__(self):
        exp = self.conn.get_server_version()
        self.assertEqual(exp, self.cnx.mysql_version)

    def test_signal(self):
        from django.db import connection

        def conn_setup(*args, **kwargs):
            conn = kwargs["connection"]
            settings.DEBUG = True
            cur = conn.cursor()
            settings.DEBUG = False
            cur.execute("SET @xyz=10")
            cur.close()

        connection_created.connect(conn_setup)
        cursor = connection.cursor()
        cursor.execute("SELECT @xyz")

        self.assertEqual((10,), cursor.fetchone())
        cursor.close()
        self.cnx.close()

    def count_conn(self, *args, **kwargs):
        try:
            self.connections += 1
        except AttributeError:
            self.connection = 1

    def test_connections(self):
        connection_created.connect(self.count_conn)
        self.connections = 0

        # Checking if DatabaseWrapper object creates a connection by default
        conn = DatabaseWrapper(settings.DATABASES["default"])
        dbo = DatabaseOperations(conn)
        dbo.adapt_timefield_value(datetime.time(3, 3, 3))
        self.assertEqual(self.connections, 0)


class DjangoDatabaseOperations(tests.MySQLConnectorTests):
    """Test the Django base.DatabaseOperations class"""

    def setUp(self):
        dbconfig = tests.get_mysql_config()
        dbconfig["use_pure"] = True
        self.conn = mysql.connector.connect(**dbconfig)
        self.cnx = DatabaseWrapper(settings.DATABASES["default"])
        self.dbo = DatabaseOperations(self.cnx)

    def test_value_to_db_time(self):
        value_to_db_time = self.dbo.adapt_timefield_value
        self.assertEqual(None, value_to_db_time(None))

        value = datetime.time(0, 0, 0)
        exp = self.conn.converter._time_to_mysql(value)
        self.assertEqual(exp, value_to_db_time(value))

        value = datetime.time(2, 5, 7)
        exp = self.conn.converter._time_to_mysql(value)
        self.assertEqual(exp, value_to_db_time(value))

    def test_value_to_db_datetime(self):
        value_to_db_datetime = self.dbo.adapt_datetimefield_value
        self.assertEqual(None, value_to_db_datetime(None))

        value = datetime.datetime(1, 1, 1)
        exp = self.conn.converter._datetime_to_mysql(value)
        self.assertEqual(exp, value_to_db_datetime(value))

        value = datetime.datetime(2, 5, 7, 10, 10)
        exp = self.conn.converter._datetime_to_mysql(value)
        self.assertEqual(exp, value_to_db_datetime(value))

    def test_bulk_insert_sql(self):
        num_values = 5
        fields = ["col1", "col2", "col3"]
        placeholder_rows = [["%s"] * len(fields) for _ in range(num_values)]
        exp = "VALUES {0}".format(
            ", ".join(["({0})".format(", ".join(["%s"] * len(fields)))] * num_values)
        )
        self.assertEqual(exp, self.dbo.bulk_insert_sql(fields, placeholder_rows))


class DjangoMySQLConverterTests(tests.MySQLConnectorTests):
    """Test the Django base.DjangoMySQLConverter class"""

    def test__time_to_python(self):
        value = b"10:11:12"
        django_converter = DjangoMySQLConverter()
        self.assertEqual(
            datetime.time(10, 11, 12),
            django_converter._time_to_python(value, dsc=None),
        )

    def test__datetime_to_python(self):
        value = b"1990-11-12 00:00:00"
        django_converter = DjangoMySQLConverter()
        self.assertEqual(
            datetime.datetime(1990, 11, 12, 0, 0, 0),
            django_converter._datetime_to_python(value, dsc=None),
        )

        settings.USE_TZ = True
        value = b"0000-00-00 00:00:00"
        django_converter = DjangoMySQLConverter()
        self.assertEqual(None, django_converter._datetime_to_python(value, dsc=None))
        settings.USE_TZ = False


class CustomDjangoMySQLConverter(DjangoMySQLConverter):
    """A custom Django MySQL converter."""

    def _customtype_to_python(self, value):
        try:
            return int(value)
        except ValueError:
            raise ValueError("Invalid value for customtype conversion")


class CustomDjangoMySQLConverterTests(tests.MySQLConnectorTests):
    """Test the Django custom MySQL converter class."""

    @staticmethod
    def create_connection(alias=DEFAULT_DB_ALIAS):
        db = django.db.connections.databases[alias]
        backend = load_backend(db["ENGINE"])
        return backend.DatabaseWrapper(db, alias)

    def test__time_to_python(self):
        value = b"10:11:12"
        django_converter = CustomDjangoMySQLConverter()
        self.assertEqual(
            datetime.time(10, 11, 12),
            django_converter._time_to_python(value, dsc=None),
        )

    def test__datetime_to_python(self):
        value = b"1990-11-12 00:00:00"
        django_converter = CustomDjangoMySQLConverter()
        self.assertEqual(
            datetime.datetime(1990, 11, 12, 0, 0, 0),
            django_converter._datetime_to_python(value, dsc=None),
        )

        settings.USE_TZ = True
        value = b"0000-00-00 00:00:00"
        django_converter = DjangoMySQLConverter()
        self.assertEqual(
            None,
            django_converter._datetime_to_python(value, dsc=None),
        )
        settings.USE_TZ = False

    def test__customtype_to_python(self):
        value = b"2021"
        django_converter = CustomDjangoMySQLConverter()
        self.assertEqual(
            2021,
            django_converter._customtype_to_python(value),
        )

    def test_invalid__customtype_to_python(self):
        value = b"abc"
        django_converter = CustomDjangoMySQLConverter()
        self.assertRaises(
            ValueError,
            django_converter._customtype_to_python,
            value,
        )

    def test_invalid_converter_class(self):
        settings.DATABASES["default"]["OPTIONS"] = {
            "converter_class": MySQLConverter,
        }
        self.assertRaises(ProgrammingError, self.create_connection)
        del settings.DATABASES["default"]["OPTIONS"]

    def test_converter_class(self):
        settings.DATABASES["default"]["OPTIONS"] = {
            "converter_class": CustomDjangoMySQLConverter,
        }
        cnx = self.create_connection()
        cnx.close()
        del settings.DATABASES["default"]["OPTIONS"]


class BugOra20106629(tests.MySQLConnectorTests):
    """CONNECTOR/PYTHON DJANGO BACKEND DOESN'T SUPPORT SAFETEXT"""

    def setUp(self):
        dbconfig = tests.get_mysql_config()
        self.conn = mysql.connector.connect(**dbconfig)
        self.cnx = DatabaseWrapper(settings.DATABASES["default"])
        self.cur = self.cnx.cursor()
        self.tbl = "BugOra20106629"
        self.cur.execute(f"DROP TABLE IF EXISTS {self.tbl}", ())
        self.cur.execute(
            f"CREATE TABLE {self.tbl}(id INT PRIMARY KEY, col1 TEXT, col2 BLOB)", ()
        )

    def tearDown(self):
        self.cur.execute("DROP TABLE IF EXISTS {0}".format(self.tbl), ())

    def test_safe_string(self):
        safe_text = SafeText("dummy & safe data <html> ")
        safe_bytes = SafeText("dummy & safe data <html> ")
        self.cur.execute(
            f"INSERT INTO {self.tbl} VALUES(%s, %s, %s)",
            (1, safe_text, safe_bytes),
        )
        self.cur.execute("SELECT col1, col2 FROM {0}".format(self.tbl), ())
        self.assertEqual(self.cur.fetchall(), [(safe_text, safe_bytes.encode())])


class BugOra34467201(tests.MySQLConnectorTests):
    """BUG#34467201: Add init_command connection option

    With this patch, support for the inti_command option is added as
    part of the connection settings. This new option allows the user to
    specify a command to be executed right after the connection is
    established, that's to say, as part of the connection initialization.
    """

    var_name = "test_var"
    var_value = "BugOra34467201"

    def setUp(self):
        settings.DATABASES["default"]["OPTIONS"] = {
            "init_command": f"SET @{self.var_name}='{self.var_value}'"
        }
        dbconfig = tests.get_mysql_config()
        self.conn = mysql.connector.connect(**dbconfig)
        self.cnx = DatabaseWrapper(settings.DATABASES["default"])

    def tearDown(self):
        with self.cnx.cursor() as cur:
            cur.execute(f"SET @{self.var_name} = NULL")
        self.cnx.close()
        del settings.DATABASES["default"]["OPTIONS"]

    def test_init_command(self):
        with self.cnx.cursor() as cur:
            cur.execute(f"SELECT @{self.var_name}")
            res = cur.fetchall()
            self.assertEqual((self.var_value,), res[0])


class BugOra35755852(tests.MySQLConnectorTests):
    """BUG#35755852: Django config raise_on_warnings is ignored without isolation_level

    When the Django config option `isolation_level` is not provided, the
    value of the option `raise_on_warnings` is ignored.

    This issue was fixed by adding the missing default value in the pop()
    function used for the `isolation_level`, which raises a `KeyError` and
    breaks the assignment of the options.
    """

    def test_missing_config(self):
        settings.DATABASES["default"]["OPTIONS"] = {"raise_on_warnings": True}
        cnx = DatabaseWrapper(settings.DATABASES["default"])
        cnx_params = cnx.get_connection_params()
        self.assertTrue(cnx_params["raise_on_warnings"])
        del settings.DATABASES["default"]["OPTIONS"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2009, 2024, Oracle and/or its affiliates.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2.0, as
# published by the Free Software Foundation.
#
# This program is designed to work with certain software (including
# but not limited to OpenSSL) that is licensed under separate terms,
# as designated in a particular file or component or in included license
# documentation. The authors of MySQL hereby grant you an
# additional permission to link the program and your derivative works
# with the separately licensed software that they have either included with
# the program or referenced in the documentation.
#
# Without limiting anything contained in the foregoing, this file,
# which is part of MySQL Connector/Python, is also subject to the
# Universal FOSS Exception, version 1.0, a copy of which can be found at
# http://oss.oracle.com/licenses/universal-foss-exception.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License, version 2.0, for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA

import os
import pathlib
import re
import shutil
import sys

sys.path.insert(0, ".")

from cpydist import BuildExt, Install, InstallLib
from cpydist.bdist import DistBinary
from cpydist.bdist_solaris import DistSolaris
from cpydist.sdist import DistSource
from setuptools import Extension, find_packages, setup

try:
    from cpydist.bdist_wheel import DistWheel
except ImportError:
    DistWheel = None

METADATA_FILES = (
    "README.txt",
    "README.rst",
    "LICENSE.txt",
    "CHANGES.txt",
    "CONTRIBUTING.md",
    "SECURITY.md",
)

VERSION_TEXT = "999.0.0"
version_py = os.path.join("lib", "mysql", "connector", "version.py")
with open(version_py, "rb") as fp:
    exec(compile(fp.read(), version_py, "exec"))

COMMAND_CLASSES = {
    "bdist": DistBinary,
    "bdist_solaris": DistSolaris,
    "build_ext": BuildExt,
    "install": Install,
    "install_lib": InstallLib,
    "sdist": DistSource,
}

if DistWheel is not None:
    COMMAND_CLASSES["bdist_wheel"] = DistWheel

# C extensions
EXTENSIONS = [
    Extension(
        "_mysql_connector",
        sources=[
            "src/exceptions.c",
            "src/mysql_capi.c",
            "src/mysql_capi_conversion.c",
            "src/mysql_connector.c",
            "src/force_cpp_linkage.cc",
        ],
        include_dirs=["src/include"],
    ),
]


def main() -> None:
    setup(
        name="mysql-connector-python",
        version=VERSION_TEXT,
        description=(
            "A self-contained Python driver for communicating with MySQL "
            "servers, using an API that is compliant with the Python "
            "Database API Specification v2.0 (PEP 249)."
        ),
        long_description=get_long_description(),
        long_description_content_type="text/x-rst",
        author="Oracle and/or its affiliates",
        author_email="",
        license="GNU GPLv2 (with FOSS License Exception)",
        keywords=[
            "mysql",
            "database",
            "db",
            "connector",
            "driver",
        ],
        project_urls={
            "Homepage": "https://dev.mysql.com/doc/connector-python/en/",
            "Documentation": "https://dev.mysql.com/doc/connector-python/en/",
            "Downloads": "https://dev.mysql.com/downloads/connector/python/",
            "Release Notes": "https://dev.mysql.com/doc/relnotes/connector-python/en/",
            "Source Code": "https://github.com/mysql/mysql-connector-python",
            "Bug System": "https://bugs.mysql.com/",
            "Slack": "https://mysqlcommunity.slack.com/messages/connectors",
            "Forums": "https://forums.mysql.com/list.php?50",
            "Blog": "https://blogs.oracle.com/mysql/",
        },
        package_dir={"": "lib"},
        packages=find_packages(where="lib"),
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "License :: OSI Approved :: GNU General Public License (GPL)",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX :: Linux",
            "Operating System :: Unix",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3.13",
            "Topic :: Database",
            "Topic :: Software Development",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Typing :: Typed",
        ],
        ext_modules=EXTENSIONS,
        cmdclass=COMMAND_CLASSES,
        python_requires=">=3.9",
        extras_require={
            "dns-srv": ["dnspython==2.6.1"],
            "gssapi": ["gssapi==1.8.3"],
            "fido2": ["fido2==1.1.2"],
            "telemetry": [
                "opentelemetry-api==1.18.0",
                "opentelemetry-sdk==1.18.0",
                "opentelemetry-exporter-otlp-proto-http==1.18.0",
            ],
        },
    )


def copy_metadata_files() -> None:
    """Copy metadata files (required by MANIFEST.in) from the
    parent directory to the current directory.
    """
    for filename in METADATA_FILES:
        shutil.copy(pathlib.Path(os.getcwd(), f"../{filename}"), pathlib.Path(f"./"))


def get_long_description() -> str:
    """Extracts a long description from the README.rst file that is suited for this specific package.
    """
    with open(pathlib.Path(os.getcwd(), "../README.rst")) as file_handle:
        # The README.rst text is meant to be shared by both mysql and mysqlx packages, so after getting it we need to
        # parse it in order to remove the bits of text that are not meaningful for this package (mysql)
        long_description = file_handle.read()
    block_matches = re.finditer(
        pattern=(
            r'(?P<module_start>\.{2}\s+={2,}\s+(?P<module_tag>\<(?P<module_name>mysql|mysqlx|both)\>)(?P<repls>\s+'
            r'\[(?:(?:,\s*)?(?:repl(?:-mysql(?:x)?)?)\("(?:[^"]+)",\s*"(?:[^"]*)"\))+\])?\s+={2,})'
            r'(?P<block_text>.+?(?=\.{2}\s+={2,}))(?P<module_end>\.{2}\s+={2,}\s+\</(?P=module_name)\>\s+={2,})'
        ),
        string=long_description,
        flags=re.DOTALL)
    for block_match in block_matches:
        if block_match.group("module_name") == 'mysqlx':
            long_description = long_description.replace(block_match.group(), "")
        else:
            block_text = block_match.group("block_text")
            if block_match.group("repls"):
                repl_matches = re.finditer(pattern=r'(?P<repl_name>repl(?:-mysql(?:x)?)?)\("'
                                                   r'(?P<repl_source>[^"]+)",\s*"(?P<repl_target>[^"]*)"\)+',
                                           string=block_match.group("repls"))
                for repl_match in repl_matches:
                    repl_name = repl_match.group("repl_name")
                    repl_source = repl_match.group("repl_source")
                    repl_target = repl_match.group("repl_target")
                    if repl_target is None:
                        repl_target = ""
                    if repl_name == "repl" or repl_name.endswith("mysql"):
                        block_text = block_text.replace(repl_source, repl_target)
            long_description = long_description.replace(block_match.group(), block_text)
    return long_description


def remove_metadata_files() -> None:
    """Remove files copied by `copy_metadata_files()`"""
    for filename in METADATA_FILES:
        os.remove(pathlib.Path(f"./{filename}"))


if __name__ == "__main__":
    copy_metadata_files()

    try:
        main()
    except Exception as err:
        raise err
    finally:
        remove_metadata_files()



# Copyright (c) 2019, 2024, Oracle and/or its affiliates.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2.0, as
# published by the Free Software Foundation.
#
# This program is designed to work with certain software (including
# but not limited to OpenSSL) that is licensed under separate terms,
# as designated in a particular file or component or in included license
# documentation. The authors of MySQL hereby grant you an
# additional permission to link the program and your derivative works
# with the separately licensed software that they have either included with
# the program or referenced in the documentation.
#
# Without limiting anything contained in the foregoing, this file,
# which is part of MySQL Connector/Python, is also subject to the
# Universal FOSS Exception, version 1.0, a copy of which can be found at
# http://oss.oracle.com/licenses/universal-foss-exception.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License, version 2.0, for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA

"""Implements the Distutils command for creating Solaris packages."""


import logging
import os
import platform
import shutil
import subprocess
import tarfile
import time

from pathlib import Path

try:
    from setuptools.errors import ExecError
except ImportError:
    ExecError = Exception

try:
    from setuptools.logging import set_threshold
except ImportError:
    set_threshold = None

from . import COMMON_USER_OPTIONS, VERSION_EXTRA, VERSION_TEXT, BaseCommand
from .bdist import DistBinary as bdist
from .utils import write_info_bin, write_info_src

SOLARIS_PKGS = {"pure": os.path.join("cpydist", "data", "solaris")}
PKGINFO = (
    'PKG="{pkg}"\n'
    'NAME="MySQL Connector/Python {ver} {lic}, MySQL driver written in '
    'Python"\n'
    'VERSION="{ver}"\n'
    'ARCH="all"\n'
    'CLASSES="none"\n'
    'CATEGORY="application"\n'
    'VENDOR="ORACLE Corporation"\n'
    'PSTAMP="{tstamp}"\n'
    'EMAIL="MySQL Release Engineering <mysql-build@oss.oracle.com>"\n'
    'BASEDIR="/"\n'
)


class DistSolaris(bdist, BaseCommand):
    """Create a Solaris distribution."""

    platf_n = "-solaris"
    platf_v = platform.version().split(".")[0]
    platf_a = "sparc" if platform.processor() == "sparc" else "x86"
    description = "create a Solaris distribution"
    user_options = COMMON_USER_OPTIONS + [
        ("dist-dir=", "d", "directory to put final built distributions in"),
        (
            "platform=",
            "p",
            f"name of the platform in resulting file (default '{platf_n}')",
        ),
        (
            "platform-version=",
            "v",
            f"version of the platform in resulting file (default '{platf_v}')",
        ),
        (
            "platform-version=",
            "a",
            "architecture, i.e. 'sparc' or 'x86' in the resulting file "
            f"(default '{platf_a}')",
        ),
        (
            "trans",
            "t",
            "transform the package into data stream (default 'False')",
        ),
    ]

    def initialize_options(self):
        """Initialize the options."""
        bdist.initialize_options(self)
        BaseCommand.initialize_options(self)
        self.name = self.distribution.get_name()
        self.version = self.distribution.get_version()
        self.version_extra = f"-{VERSION_EXTRA}" if VERSION_EXTRA else ""
        self.keep_temp = None
        self.create_dmg = False
        self.dist_dir = None
        self.started_dir = os.getcwd()
        self.platform = self.platf_n
        self.platform_version = self.platf_v
        self.architecture = self.platf_a
        self.debug = False
        self.sun_pkg_name = f"{self.name}-{self.version}{self.version_extra}.pkg"
        self.dstroot = "dstroot"
        self.sign = False
        self.identity = "MySQL Connector/Python"
        self.trans = False

    def finalize_options(self):
        """Finalize the options."""
        bdist.finalize_options(self)
        BaseCommand.finalize_options(self)
        self.set_undefined_options("bdist", ("dist_dir", "dist_dir"))
        if self.debug:
            self.log.setLevel(logging.DEBUG)
            if set_threshold:
                # Set setuptools logging level to DEBUG
                set_threshold(1)

    def _prepare_pkg_base(self, template_name, data_dir, root=""):
        """Create and populate the src base directory."""
        self.log.info("-> _prepare_pkg_base()")
        self.log.info("  template_name: %s", template_name)
        self.log.info("  data_dir: %s", data_dir)
        self.log.info("  root: %s", root)

        # copy and create necessary files
        sun_dist_name = template_name.format(self.name, self.version)
        self.sun_pkg_name = f"{sun_dist_name}.pkg"
        self.log.info("  sun_pkg_name: %s", self.sun_pkg_name)

        sun_path = os.path.join(root, self.dstroot)
        self.log.info("  sun_path: %s", sun_path)
        cwd = os.path.join(os.getcwd())
        self.log.info("Current directory: %s", cwd)

        copy_file_src_dst = []

        # No special folder for GPL or commercial. Files inside the directory
        # will determine what it is.
        data_path = os.path.join(
            sun_path,
            "usr",
            "share",
            template_name.format(self.name, self.version),
        )
        self.mkpath(data_path)

        lic = "(GPL)"
        sun_pkg_info = os.path.join(sun_path, "pkginfo")
        self.log.info("sun_pkg_info path: %s", sun_pkg_info)
        with open(sun_pkg_info, "w") as f_pkg_info:
            f_pkg_info.write(
                PKGINFO.format(
                    ver=self.version,
                    lic=lic,
                    pkg=self.name,
                    tstamp=time.ctime(),
                )
            )
            f_pkg_info.close()

        data_path = os.path.join(
            sun_path,
            "usr",
            "share",
            template_name.format(self.name, self.version),
        )
        copy_file_src_dst += [
            (
                os.path.join(cwd, "README.txt"),
                os.path.join(data_path, "README.txt"),
            ),
            (
                os.path.join(cwd, "LICENSE.txt"),
                os.path.join(data_path, "LICENSE.txt"),
            ),
            (
                os.path.join(cwd, "CHANGES.txt"),
                os.path.join(data_path, "CHANGES.txt"),
            ),
            (
                os.path.join(cwd, "docs", "INFO_SRC"),
                os.path.join(data_path, "INFO_SRC"),
            ),
            (
                os.path.join(cwd, "docs", "INFO_BIN"),
                os.path.join(data_path, "INFO_BIN"),
            ),
            (
                os.path.join(cwd, "README.rst"),
                os.path.join(data_path, "README.rst"),
            ),
            (
                os.path.join(cwd, "CONTRIBUTING.md"),
                os.path.join(data_path, "CONTRIBUTING.md"),
            ),
            (
                os.path.join(cwd, "SECURITY.md"),
                os.path.join(data_path, "SECURITY.md"),
            ),
        ]

        for src, dst in copy_file_src_dst:
            shutil.copyfile(src, dst)

    def _create_pkg(self, template_name, dmg=False, sign=False, root="", identity=""):
        """Create the Solaris package using the OS dependent commands."""
        self.log.info("-> _create_pkg()")
        self.log.info("template_name: %s", template_name)
        self.log.info("identity: %s", identity)

        sun_dist_name = template_name.format(self.name, self.version)
        self.sun_pkg_name = f"{sun_dist_name}.pkg"
        sun_pkg_contents = os.path.join(self.sun_pkg_name, "Contents")

        self.log.info("sun_dist_name: %s", sun_dist_name)
        self.log.info("sun_pkg_name: %s", self.sun_pkg_name)
        self.log.info("sun_pkg_contents: %s", sun_pkg_contents)

        sun_path = os.path.join(root, self.dstroot)
        os.chdir(sun_path)
        self.log.info("Root directory for Prototype: %s", os.getcwd())

        # Creating a Prototype file, this contains a table of contents of the
        # Package, that is suitable to be used for the package creation tool.
        self.log.info(
            f"Creating Prototype file on {self.dstroot} to describe files to install"
        )

        prototype_path = "Prototype"
        proto_tmp = "Prototype_temp"

        with open(proto_tmp, "w") as f_out:
            cmd = ["pkgproto", "."]
            pkgp_p = subprocess.Popen(cmd, shell=False, stdout=f_out, stderr=f_out)
            res = pkgp_p.wait()
            if res != 0:
                self.log.error(f"pkgproto command failed with: {res}")
                raise ExecError(f"pkgproto command failed with: {res}")
            f_out.flush()

        # log Prototype contents
        self.log.info("/n>> Prototype_temp contents >>/n")
        with open(proto_tmp, "r") as f_in:
            self.log.info(f_in.readlines())
        self.log.info("/n<< Prototype_temp contents end <</n")

        # Fix Prototype file, insert pkginfo and remove Prototype
        self.log.info("Fixing folder permissions on Prototype contents")
        with open(prototype_path, "w") as f_out:
            with open(proto_tmp, "r") as f_in:
                # Add pkginfo entry at beginning of the Prototype file
                f_out.write("i pkginfo\n")
                f_out.flush()
                for line in f_in:
                    if line.startswith("f none Prototype"):
                        continue
                    elif line.startswith("f none pkginfo"):
                        continue
                    elif line.startswith("d"):
                        tokeep = line.split(" ")[:-3]
                        tokeep.extend(["?", "?", "?", "\n"])
                        f_out.write(" ".join(tokeep))
                    elif line.startswith("f"):
                        tokeep = line.split(" ")[:-2]
                        tokeep.extend(["root", "bin", "\n"])
                        f_out.write(" ".join(tokeep))
                    else:
                        f_out.write(line)
                f_out.flush()

        # log Prototype contents
        self.log.info("/n>> Prototype contents >>/n")
        with open(prototype_path, "r") as f_in:
            self.log.info(f_in.readlines())
        self.log.info("/n<< Prototype contents end <</n")

        # Create Solaris package running the package creation command pkgmk
        self.log.info("Creating package with pkgmk")

        self.log.info("Root directory for pkgmk: %s", os.getcwd())
        self.spawn(["pkgmk", "-o", "-r", ".", "-d", "../", "-f", prototype_path])
        os.chdir("../")
        if self.debug:
            self.log.info("current directory: %s", os.getcwd())

        # gzip the package folder
        self.log.info("creating tarball")

        archive_name = f"{self.sun_pkg_name}.tar.gz"
        self.log.info("Creating tar archive '%s'", archive_name)
        with tarfile.open(archive_name, "w|gz") as tar:
            tar.add(self.name)

        if self.trans:
            self.log.info("Transforming package into data stream with pkgtrans")
            self.log.info("Current directory: %s", os.getcwd())
            self.spawn(
                [
                    "pkgtrans",
                    "-s",
                    os.getcwd(),
                    os.path.join(os.getcwd(), self.sun_pkg_name),
                    self.name,
                ]
            )

        for base, _, files in os.walk(os.getcwd()):
            for filename in files:
                if filename.endswith(".gz") or filename.endswith(".pkg"):
                    new_name = filename.replace(
                        f"{self.version}",
                        f"{self.version}{self.version_extra}{self.platform}"
                        f"{self.platform_version}-{self.architecture}",
                    )
                    file_path = os.path.join(base, filename)
                    file_dest = os.path.join(self.started_dir, self.dist_dir, new_name)
                    shutil.copyfile(file_path, file_dest)
            break

    def run(self):
        """Run the command."""
        self.mkpath(self.dist_dir)
        self.debug = self.verbose

        self.log.info("Generating INFO_SRC and INFO_BIN files")
        write_info_src(VERSION_TEXT)
        write_info_bin()

        cmd_build = self.get_finalized_command("build")
        build_base = os.path.abspath(cmd_build.build_base)
        metadata_name = self.distribution.metadata.name

        data_dir = SOLARIS_PKGS["pure"]
        sun_root = os.path.join(build_base, "sun_pure")
        cmd_install = self.reinitialize_command("install", reinit_subcommands=1)
        cmd_install.byte_code_only = self.byte_code_only
        cmd_install.compile = self.byte_code_only
        cmd_install.distribution.metadata.name = metadata_name
        cmd_install.with_mysql_capi = None
        cmd_install.root = os.path.join(sun_root, self.dstroot)
        cmd_install.ensure_finalized()
        cmd_install.run()

        template_name = ["{}"]
        if self.label:
            template_name.append(f"-{self.label}")
        template_name.append("-{}")

        self._prepare_pkg_base("".join(template_name), data_dir, root=sun_root)
        self._create_pkg(
            "".join(template_name),
            dmg=self.create_dmg,
            root=sun_root,
            sign=self.sign,
            identity=self.identity,
        )

        os.chdir(self.started_dir)

        self.remove_temp()



# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2.0, as
# published by the Free Software Foundation.
#
# This program is designed to work with certain software (including
# but not limited to OpenSSL) that is licensed under separate terms,
# as designated in a particular file or component or in included license
# documentation. The authors of MySQL hereby grant you an
# additional permission to link the program and your derivative works
# with the separately licensed software that they have either included with
# the program or referenced in the documentation.
#
# Without limiting anything contained in the foregoing, this file,
# which is part of MySQL Connector/Python, is also subject to the
# Universal FOSS Exception, version 1.0, a copy of which can be found at
# http://oss.oracle.com/licenses/universal-foss-exception.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License, version 2.0, for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA

# mypy: disable-error-code="assignment,arg-type,attr-defined,index,override,call-overload"

"""Implementation of cursor classes in pure Python."""

__all__ = ["MySQLCursor"]

import re
import warnings

from collections import namedtuple
from decimal import Decimal
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    Iterator,
    List,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from .._scripting import split_multi_statement
from ..constants import ServerFlag
from ..cursor import (
    MAX_RESULTS,
    RE_PY_MAPPING_PARAM,
    RE_PY_PARAM,
    RE_SQL_COMMENT,
    RE_SQL_FIND_PARAM,
    RE_SQL_INSERT_STMT,
    RE_SQL_INSERT_VALUES,
    RE_SQL_ON_DUPLICATE,
    RE_SQL_PYTHON_CAPTURE_PARAM_NAME,
    RE_SQL_PYTHON_REPLACE_PARAM,
)
from ..errors import (
    Error,
    InterfaceError,
    NotSupportedError,
    ProgrammingError,
    get_mysql_exception,
)
from ..types import (
    DescriptionType,
    EofPacketType,
    ParamsDictType,
    ParamsSequenceOrDictType,
    ParamsSequenceType,
    ResultType,
    RowItemType,
    RowType,
    StrOrBytes,
    WarningType,
)
from .abstracts import NAMED_TUPLE_CACHE, MySQLConnectionAbstract, MySQLCursorAbstract
from .utils import deprecated

ERR_NO_RESULT_TO_FETCH = "No result set to fetch from"


class _ParamSubstitutor:
    """Substitute parameters into a SQL statement."""

    def __init__(self, params: Sequence[bytes]) -> None:
        self.params: Sequence[bytes] = params
        self.index: int = 0

    def __call__(self, matchobj: re.Match) -> bytes:
        index = self.index
        self.index += 1
        try:
            return bytes(self.params[index])
        except IndexError:
            raise ProgrammingError(
                "Not enough parameters for the SQL statement"
            ) from None

    @property
    def remaining(self) -> int:
        """Return the number of parameters remaining to be substituted."""
        return len(self.params) - self.index


def _bytestr_format_dict(bytestr: bytes, value_dict: Dict[bytes, bytes]) -> bytes:
    """
    >>> _bytestr_format_dict(b'%(a)s', {b'a': b'foobar'})
    b'foobar
    >>> _bytestr_format_dict(b'%%(a)s', {b'a': b'foobar'})
    b'%%(a)s'
    >>> _bytestr_format_dict(b'%%%(a)s', {b'a': b'foobar'})
    b'%%foobar'
    >>> _bytestr_format_dict(b'%(x)s %(y)s',
    ...                      {b'x': b'x=%(y)s', b'y': b'y=%(x)s'})
    b'x=%(y)s y=%(x)s'
    """

    def replace(matchobj: re.Match) -> bytes:
        """Replace pattern."""
        value: Optional[bytes] = None
        groups = matchobj.groupdict()
        if groups["conversion_type"] == b"%":
            value = b"%"
        if groups["conversion_type"] == b"s":
            key = groups["mapping_key"]
            value = value_dict[key]
        if value is None:
            raise ValueError(
                f"Unsupported conversion_type: {groups['conversion_type']}"
            )
        return value

    stmt = RE_PY_MAPPING_PARAM.sub(replace, bytestr)
    return stmt


class MySQLCursor(MySQLCursorAbstract):
    """Default cursor for interacting with MySQL.

    This cursor will execute statements and handle the result. It will not automatically
    fetch all rows.

    MySQLCursor should be inherited whenever other functionallity is required.
    An example would to change the fetch* member functions to return dictionaries instead
    of lists of values.

    Implements the Python Database API Specification v2.0 (PEP-249).
    """

    async def __anext__(self) -> RowType:
        res = await self.fetchone()
        if res is not None:
            return res
        raise StopAsyncIteration

    async def close(self) -> bool:
        if not self._connection:
            return False
        self._connection.remove_cursor(self)
        await self._connection.handle_unread_result()
        await self._reset_result()
        self._connection = None
        return True

    async def _process_params_dict(
        self, params: ParamsDictType
    ) -> Dict[bytes, Union[bytes, Decimal]]:
        """Process query parameters given as dictionary."""
        res: Dict[bytes, Any] = {}
        try:
            sql_mode = await self._connection.get_sql_mode()
            to_mysql = self._connection.converter.to_mysql
            escape = self._connection.converter.escape
            quote = self._connection.converter.quote
            for key, value in params.items():
                conv = value
                conv = to_mysql(conv)
                conv = escape(conv, sql_mode)
                if not isinstance(value, Decimal):
                    conv = quote(conv)
                res[key.encode()] = conv
        except Exception as err:
            raise ProgrammingError(
                f"Failed processing pyformat-parameters; {err}"
            ) from err
        return res

    async def _process_params(
        self, params: ParamsSequenceType
    ) -> Tuple[Union[bytes, Decimal], ...]:
        """Process query parameters."""
        result = params[:]
        try:
            sql_mode = await self._connection.get_sql_mode()
            to_mysql = self._connection.converter.to_mysql
            escape = self._connection.converter.escape
            quote = self._connection.converter.quote
            result = [to_mysql(value) for value in result]
            result = [escape(value, sql_mode) for value in result]
            result = [
                quote(value) if not isinstance(params[i], Decimal) else value
                for i, value in enumerate(result)
            ]
        except Exception as err:
            raise ProgrammingError(
                f"Failed processing format-parameters; {err}"
            ) from err
        return tuple(result)

    async def _fetch_row(self, raw: bool = False) -> Optional[RowType]:
        """Return the next row in the result set."""
        if not self._connection.unread_result:
            return None
        row = None

        if self._nextrow == (None, None):
            (row, eof) = await self._connection.get_row(
                binary=self._binary,
                columns=self.description,
                raw=raw,
                read_timeout=self._read_timeout,
            )
        else:
            (row, eof) = self._nextrow

        if row:
            self._nextrow = await self._connection.get_row(
                binary=self._binary,
                columns=self.description,
                raw=raw,
                read_timeout=self._read_timeout,
            )
            eof = self._nextrow[1]
            if eof is not None:
                await self._handle_eof(eof)
            if self._rowcount == -1:
                self._rowcount = 1
            else:
                self._rowcount += 1
        if eof:
            await self._handle_eof(eof)

        return row

    async def _handle_result(self, result: ResultType) -> None:
        """Handle the result after a command was send.

        The result can be either an OK-packet or a dictionary containing column/eof
        information.

        Raises:
            InterfaceError: When result is not a dict() or result is invalid.
        """
        if not isinstance(result, dict):
            raise InterfaceError("Result was not a dict()")

        if "columns" in result:
            # Weak test, must be column/eof information
            self._description = result["columns"]
            self._connection.unread_result = True
            await self._handle_resultset()
        elif "affected_rows" in result:
            # Weak test, must be an OK-packet
            self._connection.unread_result = False
            await self._handle_noresultset(result)
        else:
            raise InterfaceError("Invalid result")

    async def _handle_resultset(self) -> None:
        """Handle the result set.

        This method handles the result set and is called after reading and storing
        column information in _handle_result(). For non-buffering cursors, this method
        is usually doing nothing.
        """

    async def _handle_noresultset(self, res: ResultType) -> None:
        """Handle result of execute() when there is no result set.

        Raises:
            ProgrammingError: When failing handling a non-resultset.
        """
        try:
            self._rowcount = res["affected_rows"]
            self._last_insert_id = res["insert_id"]
            self._warning_count = res["warning_count"]
        except (KeyError, TypeError) as err:
            raise ProgrammingError(f"Failed handling non-resultset; {err}") from None

        await self._handle_warnings()

    async def _handle_warnings(self) -> None:
        """Handle possible warnings after all results are consumed.

        Raises:
            Error: Also raises exceptions if raise_on_warnings is set.
        """
        if self._connection.get_warnings and self._warning_count:
            self._warnings = await self._fetch_warnings()

        if not self._warnings:
            return

        err = get_mysql_exception(
            self._warnings[0][1],
            self._warnings[0][2],
            warning=not self._connection.raise_on_warnings,
        )

        if self._connection.raise_on_warnings:
            raise err

        warnings.warn(err, stacklevel=4)

    async def _handle_eof(self, eof: EofPacketType) -> None:
        """Handle EOF packet."""
        self._connection.unread_result = False
        self._nextrow = (None, None)
        self._warning_count = eof["warning_count"]
        await self._handle_warnings()

    async def _reset_result(self, preserve_last_executed_stmt: bool = False) -> None:
        """Reset the cursor to default.

        Args:
            preserve_last_executed_stmt: If it is False, the last executed
                                         statement value is reset. Otherwise,
                                         such a value is preserved.
        """
        self._description = None
        self._warnings = None
        self._warning_count = 0
        self._stored_results = []
        self._rowcount = -1
        self._nextrow = (None, None)

        if not preserve_last_executed_stmt:
            # reset inner state related to statement execution
            self._executed = None
            self._executed_list = []
            self._stmt_partitions = None
            self._stmt_partition = None
            self._stmt_map_results = False

        await self.reset()

    def _have_unread_result(self) -> bool:
        """Check whether there is an unread result."""
        try:
            return self._connection.unread_result
        except AttributeError:
            return False

    def _check_executed(self) -> None:
        """Check if the statement has been executed.

        Raises:
            InterfaceError: If the statement has not been executed.
        """
        if self._executed is None:
            raise InterfaceError(ERR_NO_RESULT_TO_FETCH)

    async def _prepare_statement(
        self,
        operation: StrOrBytes,
        params: Union[Sequence[Any], Dict[str, Any]] = (),
    ) -> bytes:
        """Prepare SQL statement for execution.

        Converts the SQL statement to bytes and replaces the parameters in the
        placeholders.

        Raises:
            ProgrammingError: On converting to bytes, missing parameters or invalid
                              parameters type.
        """
        try:
            stmt = (
                operation
                if isinstance(operation, (bytes, bytearray))
                else operation.encode(self._connection.python_charset)
            )
        except (UnicodeDecodeError, UnicodeEncodeError) as err:
            raise ProgrammingError(str(err)) from err

        if params:
            if isinstance(params, dict):
                stmt = _bytestr_format_dict(
                    stmt, await self._process_params_dict(params)
                )
            elif isinstance(params, (list, tuple)):
                psub = _ParamSubstitutor(await self._process_params(params))
                stmt = RE_PY_PARAM.sub(psub, stmt)
                if psub.remaining != 0:
                    raise ProgrammingError(
                        "Not all parameters were used in the SQL statement"
                    )
            else:
                raise ProgrammingError(
                    f"Could not process parameters: {type(params).__name__}({params}),"
                    " it must be of type list, tuple or dict"
                )

        return stmt

    async def _fetch_warnings(self) -> Optional[List[WarningType]]:
        """Fetch warnings doing a SHOW WARNINGS."""
        result = []
        async with await self._connection.cursor(
            raw=False,
            read_timeout=self._read_timeout,
            write_timeout=self._write_timeout,
        ) as cur:
            await cur.execute("SHOW WARNINGS")
            result = await cur.fetchall()
        return result if result else None  # type: ignore[return-value]

    async def _batch_insert(
        self, operation: str, seq_params: Sequence[ParamsSequenceOrDictType]
    ) -> Optional[bytes]:
        """Implements multi row insert"""

        def remove_comments(match: re.Match) -> str:
            """Remove comments from INSERT statements.

            This function is used while removing comments from INSERT
            statements. If the matched string is a comment not enclosed
            by quotes, it returns an empty string, else the string itself.
            """
            if match.group(1):
                return ""
            return match.group(2)

        tmp = re.sub(
            RE_SQL_ON_DUPLICATE,
            "",
            re.sub(RE_SQL_COMMENT, remove_comments, operation),
        )

        matches = re.search(RE_SQL_INSERT_VALUES, tmp)
        if not matches:
            raise InterfaceError(
                "Failed rewriting statement for multi-row INSERT. Check SQL syntax"
            )
        fmt = matches.group(1).encode(self._connection.python_charset)
        values = []

        try:
            stmt = operation.encode(self._connection.python_charset)
            for params in seq_params:
                tmp = fmt
                if isinstance(params, dict):
                    tmp = _bytestr_format_dict(
                        tmp, await self._process_params_dict(params)
                    )
                else:
                    psub = _ParamSubstitutor(await self._process_params(params))
                    tmp = RE_PY_PARAM.sub(psub, tmp)
                    if psub.remaining != 0:
                        raise ProgrammingError(
                            "Not all parameters were used in the SQL statement"
                        )
                values.append(tmp)
            if fmt in stmt:
                stmt = stmt.replace(fmt, b",".join(values), 1)
                self._executed = stmt
                return stmt
            return None
        except (UnicodeDecodeError, UnicodeEncodeError) as err:
            raise ProgrammingError(str(err)) from err
        except Error:
            raise
        except Exception as err:
            raise InterfaceError(f"Failed executing the operation; {err}") from None

    def stored_results(self) -> Iterator[MySQLCursorAbstract]:
        """Returns an iterator for stored results.

        This method returns an iterator over results which are stored when callproc()
        is called. The iterator will provide MySQLCursorBuffered instances.
        """
        return iter(self._stored_results)

    async def callproc(
        self,
        procname: str,
        args: Sequence[Any] = (),
    ) -> Optional[Union[Dict[str, RowItemType], RowType]]:
        """Calls a stored procedure with the given arguments

        The arguments will be set during this session, meaning
        they will be called like  _<procname>__arg<nr> where
        <nr> is an enumeration (+1) of the arguments.

        Coding Example:
          1) Defining the Stored Routine in MySQL:
          CREATE PROCEDURE multiply(IN pFac1 INT, IN pFac2 INT, OUT pProd INT)
          BEGIN
            SET pProd := pFac1 * pFac2;
          END

          2) Executing in Python:
          args = (5, 5, 0)  # 0 is to hold pprod
          await cursor.callproc('multiply', args)
          print(await cursor.fetchone())

        For OUT and INOUT parameters the user should provide the
        type of the parameter as well. The argument should be a
        tuple with first item as the value of the parameter to pass
        and second argument the type of the argument.

        In the above example, one can call callproc method like:
          args = (5, 5, (0, 'INT'))
          await cursor.callproc('multiply', args)

        The type of the argument given in the tuple will be used by
        the MySQL CAST function to convert the values in the corresponding
        MySQL type (See CAST in MySQL Reference for more information)

        Does not return a value, but a result set will be
        available when the CALL-statement execute successfully.
        Raises exceptions when something is wrong.
        """
        if not procname or not isinstance(procname, str):
            raise ValueError("procname must be a string")

        if not isinstance(args, (tuple, list)):
            raise ValueError("args must be a sequence")

        self._stored_results = []

        results = []
        try:
            argnames = []
            argtypes = []

            # MySQL itself does support calling procedures with their full
            # name <database>.<procedure_name>. It's necessary to split
            # by '.' and grab the procedure name from procname.
            procname_abs = procname.split(".")[-1]
            if args:
                argvalues = []
                for idx, arg in enumerate(args):
                    argname = f"@_{procname_abs}_arg{idx + 1}"
                    argnames.append(argname)
                    if isinstance(arg, tuple):
                        argtypes.append(f" CAST({argname} AS {arg[1]})")
                        argvalues.append(arg[0])
                    else:
                        argtypes.append(argname)
                        argvalues.append(arg)

                placeholders = ",".join(f"{arg}=%s" for arg in argnames)
                await self.execute(f"SET {placeholders}", argvalues)

            call = f"CALL {procname}({','.join(argnames)})"

            # We disable consuming results temporary to make sure we
            # getting all results
            can_consume_results = self._connection.can_consume_results
            async for result in self._connection.cmd_query_iter(
                call,
                read_timeout=self._read_timeout,
                write_timeout=self._write_timeout,
            ):
                self._connection.can_consume_results = False
                if isinstance(self, (MySQLCursorDict, MySQLCursorBufferedDict)):
                    cursor_class = MySQLCursorBufferedDict
                elif isinstance(
                    self,
                    (MySQLCursorNamedTuple, MySQLCursorBufferedNamedTuple),
                ):
                    cursor_class = MySQLCursorBufferedNamedTuple
                elif self._raw:
                    cursor_class = MySQLCursorBufferedRaw
                else:
                    cursor_class = MySQLCursorBuffered
                # pylint: disable=protected-access

                # cursor_class = MySQLCursorBuffered
                cur = cursor_class(self._connection.get_self())
                cur._executed = f"(a result of {call})"
                await cur._handle_result(result)
                # pylint: enable=protected-access
                if cur.warnings is not None:
                    self._warnings = cur.warnings
                if "columns" in result:
                    results.append(cur)

            self._connection.can_consume_results = can_consume_results
            if argnames:
                # Create names aliases to be compatible with namedtuples
                args = [
                    f"{name} AS {alias}"
                    for name, alias in zip(
                        argtypes, [arg.lstrip("@_") for arg in argnames]
                    )
                ]
                select = f"SELECT {','.join(args)}"
                await self.execute(select)
                self._stored_results = results
                return await self.fetchone()

            self._stored_results = results
            return tuple()
        except Error:
            raise
        except Exception as err:
            raise InterfaceError(f"Failed calling stored routine; {err}") from None

    async def execute(
        self,
        operation: str,
        params: Union[Sequence[Any], Dict[str, Any]] = (),
        map_results: bool = False,
    ) -> None:
        if not self._connection:
            raise ProgrammingError("Cursor is not connected")

        if not operation:
            return None

        await self._connection.handle_unread_result()
        await self._reset_result()

        stmt = await self._prepare_statement(operation, params)

        self._stmt_partitions = split_multi_statement(
            sql_code=stmt, map_results=map_results
        )
        self._stmt_partition = next(self._stmt_partitions)
        self._stmt_map_results = map_results
        self._executed_list = self._stmt_partition["single_stmts"]
        self._executed = (
            self._stmt_partition["single_stmts"].popleft()
            if map_results
            else self._stmt_partition["mappable_stmt"]
        )

        await self._handle_result(
            await self._connection.cmd_query(
                self._stmt_partition["mappable_stmt"],
                read_timeout=self._read_timeout,
                write_timeout=self._write_timeout,
            )
        )

        return None

    @deprecated(
        "executemulti() is deprecated and will be removed in a future release. "
        + "Use execute() instead."
    )
    async def executemulti(
        self,
        operation: str,
        params: Union[Sequence[Any], Dict[str, Any]] = (),
        map_results: bool = False,
    ) -> None:
        return await self.execute(
            operation=operation, params=params, map_results=map_results
        )

    async def executemany(
        self,
        operation: str,
        seq_params: Sequence[ParamsSequenceType],
    ) -> None:
        """Prepare and execute a MySQL Prepared Statement many times.

        This method will prepare the given operation and execute with each tuple found
        the list seq_params.

        If the cursor instance already had a prepared statement, it is first closed.
        """
        if not operation or not seq_params:
            return None
        await self._connection.handle_unread_result()

        try:
            _ = iter(seq_params)
        except TypeError as err:
            raise ProgrammingError("Parameters for query must be an Iterable") from err

        # Optimize INSERTs by batching them
        if re.match(RE_SQL_INSERT_STMT, operation):
            if not seq_params:
                self._rowcount = 0
                return None
            stmt = await self._batch_insert(operation, seq_params)
            if stmt is not None:
                self._executed = stmt
                return await self.execute(stmt)

        rowcnt = 0
        try:
            for params in seq_params:
                await self.execute(operation, params)
                if self.with_rows and self._have_unread_result():
                    await self.fetchall()
                rowcnt += self._rowcount
        except (ValueError, TypeError) as err:
            raise ProgrammingError(f"Failed executing the operation; {err}") from None
        self._rowcount = rowcnt

    async def fetchone(self) -> Optional[RowType]:
        """Return next row of a query result set.

        Raises:
            InterfaceError: If there is no result to fetch.

        Returns:
            tuple or None: A row from query result set.
        """
        if self._executed is None:
            raise InterfaceError(ERR_NO_RESULT_TO_FETCH)
        return await self._fetch_row()

    async def fetchall(self) -> List[RowType]:
        """Return all rows of a query result set.

        Returns:
            list: A list of tuples with all rows of a query result set.
        """
        if self._executed is None:
            raise InterfaceError(ERR_NO_RESULT_TO_FETCH)

        if not self._connection.unread_result:
            return []

        rows, eof = await self._connection.get_rows(read_timeout=self._read_timeout)
        if self._nextrow[0]:
            rows.insert(0, self._nextrow[0])

        await self._handle_eof(eof)
        rowcount = len(rows)
        if rowcount >= 0 and self._rowcount == -1:
            self._rowcount = 0
        self._rowcount += rowcount
        return rows

    async def fetchmany(self, size: Optional[int] = None) -> List[RowType]:
        """Return the next set of rows of a query result set.

        When no more rows are available, it returns an empty list.
        The number of rows returned can be specified using the size argument, which
        defaults to one.

        Returns:
            list: The next set of rows of a query result set.
        """
        self._check_executed()
        res = []
        cnt = size or self.arraysize
        while cnt > 0:
            cnt -= 1
            row = await self.fetchone()
            if row:
                res.append(row)

        return res

    async def nextset(self) -> Optional[bool]:
        if self._connection._have_next_result:
            # prepare cursor to load the next result set, and ultimately, load it.
            await self._connection.handle_unread_result()
            await self._reset_result(preserve_last_executed_stmt=True)
            await self._handle_result(
                await self._connection._handle_result(
                    await self._connection._socket.read(read_timeout=self._read_timeout)
                )
            )

            # if mapping is enabled, run the if-block, otherwise simply return `True`.
            if self._stmt_partitions is not None and self._stmt_map_results:
                if not self._stmt_partition["single_stmts"]:
                    # It means there are still results to be consumed, but no more
                    # statements to relate these results to.
                    # In this case, we raise a no fatal error and don't clear
                    # `_executed` so its current value is reported when users
                    # access the property `statement`.
                    # If this case ever happens, a bug report should be filed,
                    # assuming it is happening on supported use cases.
                    warnings.warn(
                        "MappingWarning: Number of result sets greater than number "
                        "of single statements."
                    )
                else:
                    self._executed = self._stmt_partition["single_stmts"].popleft()
            return True
        if self._stmt_partitions is not None:
            # Let's see if there are more mappable statements (partitions)
            # to be executed.
            # If there are no more partitions, we simply return `None`, otherwise
            # we execute the correponding mappable multi statement and repeat the
            # process all over again.
            try:
                self._stmt_partition = next(self._stmt_partitions)
            except StopIteration:
                pass
            else:
                # This block only happens when mapping is enabled because when it
                # is disabled, only one partition is generated, and at this point,
                # such partiton has already been processed.
                self._executed = self._stmt_partition["single_stmts"].popleft()
                await self._handle_result(
                    await self._connection.cmd_query(
                        self._stmt_partition["mappable_stmt"],
                        read_timeout=self._read_timeout,
                        write_timeout=self._write_timeout,
                    )
                )
                return True

        await self._reset_result()
        return None


class MySQLCursorBuffered(MySQLCursor):
    """Cursor which fetches rows within execute()."""

    def __init__(
        self,
        connection: MySQLConnectionAbstract,
        read_timeout: Optional[int] = None,
        write_timeout: Optional[int] = None,
    ) -> None:
        super().__init__(connection, read_timeout, write_timeout)
        self._rows: Optional[List[RowType]] = None
        self._next_row: int = 0

    async def _handle_resultset(self) -> None:
        """Handle the result set.

        This method handles the result set and is called after reading and storing
        column information in _handle_result(). For non-buffering cursors, this method
        is usually doing nothing.
        """
        self._rows, eof = await self._connection.get_rows(
            raw=self._raw, read_timeout=self._read_timeout
        )
        self._rowcount = len(self._rows)
        await self._handle_eof(eof)
        self._next_row = 0
        self._connection.unread_result = False

    async def _fetch_row(self, raw: bool = False) -> Optional[RowType]:
        """Return the next row in the result set."""
        row = None
        try:
            row = self._rows[self._next_row]
        except (IndexError, TypeError):
            return None
        self._next_row += 1
        return row

    async def reset(self, free: bool = True) -> None:
        """Reset the cursor to default."""
        self._rows = None

    async def fetchone(self) -> Optional[RowType]:
        """Return next row of a query result set.

        Returns:
            tuple or None: A row from query result set.
        """
        self._check_executed()
        return await self._fetch_row()

    async def fetchall(self) -> List[RowType]:
        """Return all rows of a query result set.

        Returns:
            list: A list of tuples with all rows of a query result set.
        """
        if self._executed is None:
            raise InterfaceError(ERR_NO_RESULT_TO_FETCH)
        if self._rows is None:
            return []
        res = []
        res = self._rows[self._next_row :]
        self._next_row = len(self._rows)
        return res

    @property
    def with_rows(self) -> bool:
        return self._rows is not None


class MySQLCursorRaw(MySQLCursor):
    """Skip conversion from MySQL datatypes to Python types when fetching rows."""

    def __init__(
        self,
        connection: MySQLConnectionAbstract,
        read_timeout: Optional[int] = None,
        write_timeout: Optional[int] = None,
    ) -> None:
        super().__init__(connection, read_timeout, write_timeout)
        self._raw: bool = True

    async def fetchone(self) -> Optional[RowType]:
        """Return next row of a query result set.

        Returns:
            tuple or None: A row from query result set.
        """
        self._check_executed()
        return await self._fetch_row(raw=True)

    async def fetchall(self) -> List[RowType]:
        """Return all rows of a query result set.

        Returns:
            list: A list of tuples with all rows of a query result set.
        """
        self._check_executed()
        if not self._have_unread_result():
            return []
        rows, eof = await self._connection.get_rows(
            raw=True, read_timeout=self._read_timeout
        )
        if self._nextrow[0]:
            rows.insert(0, self._nextrow[0])
        await self._handle_eof(eof)
        rowcount = len(rows)
        if rowcount >= 0 and self._rowcount == -1:
            self._rowcount = 0
        self._rowcount += rowcount
        return rows


class MySQLCursorBufferedRaw(MySQLCursorBuffered):
    """
    Cursor which skips conversion from MySQL datatypes to Python types when
    fetching rows and fetches rows within execute().
    """

    def __init__(
        self,
        connection: MySQLConnectionAbstract,
        read_timeout: Optional[int] = None,
        write_timeout: Optional[int] = None,
    ) -> None:
        super().__init__(connection, read_timeout, write_timeout)
        self._raw: bool = True

    @property
    def with_rows(self) -> bool:
        return self._rows is not None


class MySQLCursorDict(MySQLCursor):
    """
    Cursor fetching rows as dictionaries.

    The fetch methods of this class will return dictionaries instead of tuples.
    Each row is a dictionary that looks like:
        row = {
            "col1": value1,
            "col2": value2
        }
    """

    def _row_to_python(
        self,
        rowdata: RowType,
        desc: Optional[List[DescriptionType]] = None,  # pylint: disable=unused-argument
    ) -> Optional[Dict[str, RowItemType]]:
        """Convert a MySQL text result row to Python types

        Returns a dictionary.
        """
        return dict(zip(self.column_names, rowdata)) if rowdata else None

    async def fetchone(self) -> Optional[Dict[str, RowItemType]]:
        """Return next row of a query result set.

        Returns:
            dict or None: A dict from query result set.
        """
        return self._row_to_python(await super().fetchone(), self.description)

    async def fetchall(self) -> List[Optional[Dict[str, RowItemType]]]:
        """Return all rows of a query result set.

        Returns:
            list: A list of dictionaries with all rows of a query
                  result set where column names are used as keys.
        """
        return [
            self._row_to_python(row, self.description)
            for row in await super().fetchall()
            if row
        ]


class MySQLCursorNamedTuple(MySQLCursor):
    """
    Cursor fetching rows as named tuple.

    The fetch methods of this class will return namedtuples instead of tuples.
    Each row is returned as a namedtuple and the values can be accessed as:
    row.col1, row.col2
    """

    def _row_to_python(
        self,
        rowdata: RowType,
        desc: Optional[List[DescriptionType]] = None,  # pylint: disable=unused-argument
    ) -> Optional[RowType]:
        """Convert a MySQL text result row to Python types

        Returns a named tuple.
        """
        row = rowdata

        if row:
            columns = tuple(self.column_names)
            try:
                named_tuple = NAMED_TUPLE_CACHE[columns]
            except KeyError:
                named_tuple = namedtuple("Row", columns)  # type:ignore[no-redef, misc]
                NAMED_TUPLE_CACHE[columns] = named_tuple
            return named_tuple(*row)
        return None

    async def fetchone(self) -> Optional[RowType]:
        """Return next row of a query result set.

        Returns:
            tuple or None: A row from query result set.
        """
        row = await super().fetchone()
        if not row:
            return None
        return (
            self._row_to_python(row, self.description)
            if hasattr(self._connection, "converter")
            else row
        )

    async def fetchall(self) -> List[Optional[RowType]]:
        """Return all rows of a query result set.

        Returns:
            list: A list of tuples with all rows of a query result set.
        """
        return [
            self._row_to_python(row, self.description)
            for row in await super().fetchall()
            if row
        ]


class MySQLCursorBufferedDict(MySQLCursorDict, MySQLCursorBuffered):
    """
    Buffered Cursor fetching rows as dictionaries.
    """

    async def fetchone(self) -> Optional[Dict[str, RowItemType]]:
        """Return next row of a query result set.

        Returns:
            tuple or None: A row from query result set.
        """
        self._check_executed()
        row = await self._fetch_row()
        if row:
            return self._row_to_python(row, self.description)
        return None

    async def fetchall(self) -> List[Optional[Dict[str, RowItemType]]]:
        """Return all rows of a query result set.

        Returns:
            list: A list of tuples with all rows of a query result set.
        """
        if self._executed is None or self._rows is None:
            raise InterfaceError(ERR_NO_RESULT_TO_FETCH)
        res = []
        for row in self._rows[self._next_row :]:
            res.append(self._row_to_python(row, self.description))
        self._next_row = len(self._rows)
        return res


class MySQLCursorBufferedNamedTuple(MySQLCursorNamedTuple, MySQLCursorBuffered):
    """
    Buffered Cursor fetching rows as named tuple.
    """

    async def fetchone(self) -> Optional[RowType]:
        """Return next row of a query result set.

        Returns:
            tuple or None: A row from query result set.
        """
        self._check_executed()
        row = await self._fetch_row()
        if row:
            return self._row_to_python(row, self.description)
        return None

    async def fetchall(self) -> List[Optional[RowType]]:
        """Return all rows of a query result set.

        Returns:
            list: A list of tuples with all rows of a query result set.
        """
        if self._executed is None or self._rows is None:
            raise InterfaceError(ERR_NO_RESULT_TO_FETCH)
        res = []
        for row in self._rows[self._next_row :]:
            res.append(self._row_to_python(row, self.description))
        self._next_row = len(self._rows)
        return res


class MySQLCursorPrepared(MySQLCursor):
    """Cursor using MySQL Prepared Statements"""

    def __init__(
        self,
        connection: MySQLConnectionAbstract,
        read_timeout: Optional[int] = None,
        write_timeout: Optional[int] = None,
    ):
        super().__init__(connection, read_timeout, write_timeout)
        self._rows: Optional[List[RowType]] = None
        self._next_row: int = 0
        self._prepared: Optional[Dict[str, Union[int, List[DescriptionType]]]] = None
        self._binary: bool = True
        self._have_result: Optional[bool] = None
        self._last_row_sent: bool = False
        self._cursor_exists: bool = False

    async def reset(self, free: bool = True) -> None:
        if self._prepared:
            try:
                await self._connection.cmd_stmt_close(
                    self._prepared["statement_id"],
                    read_timeout=self._read_timeout,
                    write_timeout=self._write_timeout,
                )
            except Error:
                # We tried to deallocate, but it's OK when we fail.
                pass
            self._prepared = None
        self._last_row_sent = False
        self._cursor_exists = False

    async def _handle_noresultset(self, res: ResultType) -> None:
        self._handle_server_status(
            res.get("status_flag", res.get("server_status", 0)),
        )
        await super()._handle_noresultset(res)

    def _handle_server_status(self, flags: int) -> None:
        self._cursor_exists = flags & ServerFlag.STATUS_CURSOR_EXISTS != 0
        self._last_row_sent = flags & ServerFlag.STATUS_LAST_ROW_SENT != 0

    async def _handle_eof(self, eof: EofPacketType) -> None:
        self._handle_server_status(
            eof.get("status_flag", eof.get("server_status", 0)),
        )
        await super()._handle_eof(eof)

    async def callproc(self, procname: Any, args: Any = ()) -> NoReturn:
        """Calls a stored procedue

        Not supported with MySQLCursorPrepared.
        """
        raise NotSupportedError()

    @deprecated(
        "executemulti() is deprecated and will be removed in a future release. "
        + "Use execute() instead."
    )
    async def executemulti(
        self,
        operation: str,
        params: Union[Sequence[Any], Dict[str, Any]] = (),
        map_results: bool = False,
    ) -> AsyncGenerator[MySQLCursorAbstract, None]:
        """Execute multiple statements.

        Executes the given operation substituting any markers with the given
        parameters.
        """
        raise NotSupportedError()

    async def close(self) -> None:
        """Close the cursor

        This method will try to deallocate the prepared statement and close
        the cursor.
        """
        await self.reset()
        await super().close()

    def _row_to_python(self, rowdata: Any, desc: Any = None) -> Any:
        """Convert row data from MySQL to Python types

        The conversion is done while reading binary data in the protocol module.
        """

    async def _handle_result(self, result: ResultType) -> None:
        """Handle result after execution"""
        if isinstance(result, dict):
            self._connection.unread_result = False
            self._have_result = False
            await self._handle_noresultset(result)
        else:
            self._description = result[1]
            self._connection.unread_result = True
            self._have_result = True
            if "status_flag" in result[2]:  # type: ignore[operator]
                self._handle_server_status(result[2]["status_flag"])
            elif "server_status" in result[2]:  # type: ignore[operator]
                self._handle_server_status(result[2]["server_status"])

    async def execute(
        self,
        operation: StrOrBytes,
        params: Optional[ParamsSequenceOrDictType] = None,
        map_results: bool = False,
    ) -> None:
        """Prepare and execute a MySQL Prepared Statement

        This method will prepare the given operation and execute it using
        the optionally given parameters.

        If the cursor instance already had a prepared statement, it is
        first closed.

        *Argument "map_results" is unused as multi statement execution
        is not supported for prepared statements*.

        Raises:
            ProgrammingError: When providing a multi statement operation
                              or setting *map_results* to True.
        """
        if map_results:
            raise ProgrammingError(
                "Multi statement execution not supported for prepared statements."
            )

        if not self._connection:
            raise ProgrammingError("Cursor is not connected")

        if not operation:
            return None

        charset = self._connection.charset
        if charset == "utf8mb4":
            charset = "utf8"

        if not isinstance(operation, str):
            try:
                operation = operation.decode(charset)
            except UnicodeDecodeError as err:
                raise ProgrammingError(str(err)) from err

        if isinstance(params, dict):
            replacement_keys = re.findall(RE_SQL_PYTHON_CAPTURE_PARAM_NAME, operation)
            try:
                # Replace params dict with params tuple in correct order
                params = tuple(params[key] for key in replacement_keys)
            except KeyError as err:
                raise ProgrammingError(
                    "Not all placeholders were found in the parameters dict"
                ) from err
            # Convert %(name)s to ? before sending it to MySQL
            operation = re.sub(RE_SQL_PYTHON_REPLACE_PARAM, "?", operation)

        if operation is not self._executed:
            if self._prepared:
                await self._connection.cmd_stmt_close(
                    self._prepared["statement_id"],
                    read_timeout=self._read_timeout,
                    write_timeout=self._write_timeout,
                )
            self._executed = operation

            try:
                operation = operation.encode(charset)
            except UnicodeEncodeError as err:
                raise ProgrammingError(str(err)) from err

            if b"%s" in operation:
                # Convert %s to ? before sending it to MySQL
                operation = re.sub(RE_SQL_FIND_PARAM, b"?", operation)

            try:
                self._prepared = await self._connection.cmd_stmt_prepare(
                    operation,
                    read_timeout=self._read_timeout,
                    write_timeout=self._write_timeout,
                )
            except Error:
                self._executed = None
                raise

        await self._connection.cmd_stmt_reset(
            self._prepared["statement_id"],
            read_timeout=self._read_timeout,
            write_timeout=self._write_timeout,
        )

        if self._prepared["parameters"] and not params:
            return
        if params:
            if not isinstance(params, (tuple, list)):
                raise ProgrammingError(
                    errno=1210,
                    msg="Incorrect type of argument: "
                    f"{type(params).__name__}({params})"
                    ", it must be of type tuple or list the argument given to "
                    "the prepared statement",
                )
            if len(self._prepared["parameters"]) != len(params):
                raise ProgrammingError(
                    errno=1210,
                    msg="Incorrect number of arguments executing prepared statement",
                )

        if params is None:
            params = ()
        res = await self._connection.cmd_stmt_execute(
            self._prepared["statement_id"],
            data=params,
            parameters=self._prepared["parameters"],
            read_timeout=self._read_timeout,
            write_timeout=self._write_timeout,
        )
        await self._handle_result(res)

    async def executemany(
        self,
        operation: str,
        seq_params: Sequence[ParamsSequenceType],
    ) -> None:
        """Prepare and execute a MySQL Prepared Statement many times

        This method will prepare the given operation and execute with each
        tuple found the list seq_params.

        If the cursor instance already had a prepared statement, it is
        first closed.

        executemany() simply calls execute().
        """
        if not operation or not seq_params:
            return None
        await self._connection.handle_unread_result()

        try:
            _ = iter(seq_params)
        except TypeError as err:
            raise ProgrammingError("Parameters for query must be an Iterable") from err

        rowcnt = 0
        try:
            for params in seq_params:
                await self.execute(operation, params)
                if self.with_rows and self._have_unread_result():
                    await self.fetchall()
                rowcnt += self._rowcount
        except (ValueError, TypeError) as err:
            raise InterfaceError(f"Failed executing the operation; {err}") from None
        self._rowcount = rowcnt

    async def fetchone(self) -> Optional[RowType]:
        """Return next row of a query result set.

        Returns:
            tuple or None: A row from query result set.
        """
        self._check_executed()
        if self._cursor_exists:
            await self._connection.cmd_stmt_fetch(
                self._prepared["statement_id"],
                read_timeout=self._read_timeout,
                write_timeout=self._write_timeout,
            )
        return await self._fetch_row() or None

    async def fetchmany(self, size: Optional[int] = None) -> List[RowType]:
        """Return the next set of rows of a query result set.

        When no more rows are available, it returns an empty list.
        The number of rows returned can be specified using the size argument,
        which defaults to one.

        Returns:
            list: The next set of rows of a query result set.
        """
        self._check_executed()
        res = []
        cnt = size or self.arraysize
        while cnt > 0 and self._have_unread_result():
            cnt -= 1
            row = await self._fetch_row()
            if row:
                res.append(row)
        return res

    async def fetchall(self) -> List[RowType]:
        """Return all rows of a query result set.

        Returns:
            list: A list of tuples with all rows of a query result set.
        """
        self._check_executed()
        rows = []
        if self._nextrow[0]:
            rows.append(self._nextrow[0])
        while self._have_unread_result():
            if self._cursor_exists:
                await self._connection.cmd_stmt_fetch(
                    self._prepared["statement_id"],
                    MAX_RESULTS,
                    read_timeout=self._read_timeout,
                    write_timeout=self._write_timeout,
                )
            tmp, eof = await self._connection.get_rows(
                binary=self._binary,
                columns=self.description,
                read_timeout=self._read_timeout,
            )
            rows.extend(tmp)
            await self._handle_eof(eof)
        self._rowcount = len(rows)
        return rows


class MySQLCursorPreparedDict(MySQLCursorDict, MySQLCursorPrepared):  # type: ignore[misc]
    """
    This class is a blend of features from MySQLCursorDict and MySQLCursorPrepared

    Multiple inheritance in python is allowed but care must be taken
    when assuming methods resolution. In the case of multiple
    inheritance, a given attribute is first searched in the current
    class if it's not found then it's searched in the parent classes.
    The parent classes are searched in a left-right fashion and each
    class is searched once.
    Based on python's attribute resolution, in this case, attributes
    are searched as follows:
    1. MySQLCursorPreparedDict (current class)
    2. MySQLCursorDict (left parent class)
    3. MySQLCursorPrepared (right parent class)
    4. MySQLCursor (base class)
    """

    async def fetchmany(
        self, size: Optional[int] = None
    ) -> List[Dict[str, RowItemType]]:
        """Return the next set of rows of a query result set.

        When no more rows are available, it returns an empty list.
        The number of rows returned can be specified using the size argument,
        which defaults to one.

        Returns:
            list: The next set of rows of a query result set represented
                  as a list of dictionaries where column names are used as keys.
        """
        return [
            self._row_to_python(row, self.description)
            for row in await super().fetchmany(size=size)
            if row
        ]


class MySQLCursorPreparedNamedTuple(MySQLCursorNamedTuple, MySQLCursorPrepared):
    """
    This class is a blend of features from MySQLCursorNamedTuple and MySQLCursorPrepared
    """

    async def fetchmany(self, size: Optional[int] = None) -> List[RowType]:
        """Return the next set of rows of a query result set.

        When no more rows are available, it returns an empty list.
        The number of rows returned can be specified using the size argument,
        which defaults to one.

        Returns:
            list: The next set of rows of a query result set represented
                  as a list of named tuples where column names are used as names.
        """
        return [
            self._row_to_python(row, self.description)
            for row in await super().fetchmany(size=size)
            if row
        ]


class MySQLCursorPreparedRaw(MySQLCursorPrepared):
    """
    This class is a blend of features from MySQLCursorRaw and MySQLCursorPrepared
    """

    def __init__(
        self,
        connection: MySQLConnectionAbstract,
        read_timeout: Optional[int] = None,
        write_timeout: Optional[int] = None,
    ):
        super().__init__(connection, read_timeout, write_timeout)
        self._raw: bool = True

    async def fetchone(self) -> Optional[RowType]:
        """Return next row of a query result set.

        Returns:
            tuple or None: A row from query result set.
        """
        self._check_executed()
        if self._cursor_exists:
            await self._connection.cmd_stmt_fetch(
                self._prepared["statement_id"],
                read_timeout=self._read_timeout,
                write_timeout=self._write_timeout,
            )
        return await self._fetch_row(raw=self._raw) or None

    async def fetchmany(self, size: Optional[int] = None) -> List[RowType]:
        """Return the next set of rows of a query result set.

        When no more rows are available, it returns an empty list.
        The number of rows returned can be specified using the size argument,
        which defaults to one.

        Returns:
            list: The next set of rows of a query result set.
        """
        self._check_executed()
        res = []
        cnt = size or self.arraysize
        while cnt > 0 and self._have_unread_result():
            cnt -= 1
            row = await self._fetch_row(raw=self._raw)
            if row:
                res.append(row)
        return res

    async def fetchall(self) -> List[RowType]:
        """Return all rows of a query result set.

        Returns:
            list: A list of tuples with all rows of a query result set.
        """
        self._check_executed()
        rows = []
        if self._nextrow[0]:
            rows.append(self._nextrow[0])
        while self._have_unread_result():
            if self._cursor_exists:
                await self._connection.cmd_stmt_fetch(
                    self._prepared["statement_id"],
                    MAX_RESULTS,
                    read_timeout=self._read_timeout,
                    write_timeout=self._write_timeout,
                )
            tmp, eof = await self._connection.get_rows(
                raw=self._raw,
                binary=self._binary,
                columns=self.description,
                read_timeout=self._read_timeout,
            )
            rows.extend(tmp)
            await self._handle_eof(eof)
        self._rowcount = len(rows)
        return rows



# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2.0, as
# published by the Free Software Foundation.
#
# This program is designed to work with certain software (including
# but not limited to OpenSSL) that is licensed under separate terms,
# as designated in a particular file or component or in included license
# documentation. The authors of MySQL hereby grant you an
# additional permission to link the program and your derivative works
# with the separately licensed software that they have either included with
# the program or referenced in the documentation.
#
# Without limiting anything contained in the foregoing, this file,
# which is part of MySQL Connector/Python, is also subject to the
# Universal FOSS Exception, version 1.0, a copy of which can be found at
# http://oss.oracle.com/licenses/universal-foss-exception.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License, version 2.0, for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA

"""MySQL Connector/Python - MySQL driver written in Python."""

__all__ = ["CMySQLConnection", "MySQLConnection", "connect"]

import random

from typing import Any

from ..constants import DEFAULT_CONFIGURATION
from ..errors import Error, InterfaceError, ProgrammingError
from ..pooling import ERROR_NO_CEXT
from .abstracts import MySQLConnectionAbstract
from .connection import MySQLConnection

try:
    import dns.exception
    import dns.resolver
except ImportError:
    HAVE_DNSPYTHON = False
else:
    HAVE_DNSPYTHON = True


try:
    from .connection_cext import CMySQLConnection
except ImportError:
    CMySQLConnection = None


async def connect(*args: Any, **kwargs: Any) -> MySQLConnectionAbstract:
    """Creates or gets a MySQL connection object.

    In its simpliest form, `connect()` will open a connection to a
    MySQL server and return a `MySQLConnectionAbstract` subclass
    object such as `MySQLConnection` or `CMySQLConnection`.

    When any connection pooling arguments are given, for example `pool_name`
    or `pool_size`, a pool is created or a previously one is used to return
    a `PooledMySQLConnection`.

    Args:
        *args: N/A.
        **kwargs: For a complete list of possible arguments, see [1]. If no arguments
                  are given, it uses the already configured or default values.

    Returns:
        A `MySQLConnectionAbstract` subclass instance (such as `MySQLConnection` or
        a `CMySQLConnection`) instance.

    Examples:
        A connection with the MySQL server can be established using either the
        `mysql.connector.connect()` method or a `MySQLConnectionAbstract` subclass:
        ```
        >>> from mysql.connector.aio import MySQLConnection, HAVE_CEXT
        >>>
        >>> cnx1 = await mysql.connector.aio.connect(user='joe', database='test')
        >>> cnx2 = MySQLConnection(user='joe', database='test')
        >>> await cnx2.connect()
        >>>
        >>> cnx3 = None
        >>> if HAVE_CEXT:
        >>>     from mysql.connector.aio import CMySQLConnection
        >>>     cnx3 = CMySQLConnection(user='joe', database='test')
        ```

    References:
        [1]: https://dev.mysql.com/doc/connector-python/en/connector-python-connectargs.html
    """
    # DNS SRV
    dns_srv = kwargs.pop("dns_srv") if "dns_srv" in kwargs else False

    if not isinstance(dns_srv, bool):
        raise InterfaceError("The value of 'dns-srv' must be a boolean")

    if dns_srv:
        if not HAVE_DNSPYTHON:
            raise InterfaceError(
                "MySQL host configuration requested DNS "
                "SRV. This requires the Python dnspython "
                "module. Please refer to documentation"
            )
        if "unix_socket" in kwargs:
            raise InterfaceError(
                "Using Unix domain sockets with DNS SRV lookup is not allowed"
            )
        if "port" in kwargs:
            raise InterfaceError(
                "Specifying a port number with DNS SRV lookup is not allowed"
            )
        if "failover" in kwargs:
            raise InterfaceError(
                "Specifying multiple hostnames with DNS SRV look up is not allowed"
            )
        if "host" not in kwargs:
            kwargs["host"] = DEFAULT_CONFIGURATION["host"]

        try:
            srv_records = dns.resolver.query(kwargs["host"], "SRV")
        except dns.exception.DNSException:
            raise InterfaceError(
                f"Unable to locate any hosts for '{kwargs['host']}'"
            ) from None

        failover = []
        for srv in srv_records:
            failover.append(
                {
                    "host": srv.target.to_text(omit_final_dot=True),
                    "port": srv.port,
                    "priority": srv.priority,
                    "weight": srv.weight,
                }
            )

        failover.sort(key=lambda x: (x["priority"], -x["weight"]))
        kwargs["failover"] = [
            {"host": srv["host"], "port": srv["port"]} for srv in failover
        ]

    # Failover
    if "failover" in kwargs:
        return await _get_failover_connection(**kwargs)

    # Use C Extension by default
    use_pure = kwargs.get("use_pure", False)
    if "use_pure" in kwargs:
        del kwargs["use_pure"]  # Remove 'use_pure' from kwargs
        if not use_pure and CMySQLConnection is None:
            raise ImportError(ERROR_NO_CEXT)

    if CMySQLConnection and not use_pure:
        cnx = CMySQLConnection(*args, **kwargs)
    else:
        cnx = MySQLConnection(*args, **kwargs)
    await cnx.connect()
    return cnx


async def _get_failover_connection(**kwargs: Any) -> MySQLConnectionAbstract:
    """Return a MySQL connection and try to failover if needed.

    An InterfaceError is raise when no MySQL is available. ValueError is
    raised when the failover server configuration contains an illegal
    connection argument. Supported arguments are user, password, host, port,
    unix_socket and database. ValueError is also raised when the failover
    argument was not provided.

    Returns MySQLConnection instance.
    """
    config = kwargs.copy()
    try:
        failover = config["failover"]
    except KeyError:
        raise ValueError("failover argument not provided") from None
    del config["failover"]

    support_cnx_args = set(
        [
            "user",
            "password",
            "host",
            "port",
            "unix_socket",
            "database",
            "pool_name",
            "pool_size",
            "priority",
        ]
    )

    # First check if we can add all use the configuration
    priority_count = 0
    for server in failover:
        diff = set(server.keys()) - support_cnx_args
        if diff:
            arg = "s" if len(diff) > 1 else ""
            lst = ", ".join(diff)
            raise ValueError(
                f"Unsupported connection argument {arg} in failover: {lst}"
            )
        if hasattr(server, "priority"):
            priority_count += 1

        server["priority"] = server.get("priority", 100)
        if server["priority"] < 0 or server["priority"] > 100:
            raise InterfaceError(
                "Priority value should be in the range of 0 to 100, "
                f"got : {server['priority']}"
            )
        if not isinstance(server["priority"], int):
            raise InterfaceError(
                "Priority value should be an integer in the range of 0 to "
                f"100, got : {server['priority']}"
            )

    if 0 < priority_count < len(failover):
        raise ProgrammingError(
            "You must either assign no priority to any "
            "of the routers or give a priority for "
            "every router"
        )

    server_directory = {}
    server_priority_list = []
    for server in sorted(failover, key=lambda x: x["priority"], reverse=True):
        if server["priority"] not in server_directory:
            server_directory[server["priority"]] = [server]
            server_priority_list.append(server["priority"])
        else:
            server_directory[server["priority"]].append(server)

    for priority in server_priority_list:
        failover_list = server_directory[priority]
        for _ in range(len(failover_list)):
            last = len(failover_list) - 1
            index = random.randint(0, last)
            server = failover_list.pop(index)
            new_config = config.copy()
            new_config.update(server)
            new_config.pop("priority", None)
            try:
                return await connect(**new_config)
            except Error:
                # If we failed to connect, we try the next server
                pass

    raise InterfaceError("Unable to connect to any of the target hosts")
